import numpy as np
from rpy2.robjects.numpy2ri import numpy2ri
from rpy2.robjects.packages import importr

R_VINECOPULA = importr('VineCopula')

def get_param1_interval(copula):
    """
    """
    assert isinstance(copula, (int, str)), \
        TypeError("Input must be int or str")
    if isinstance(copula, str):
        copula = int(R_VINECOPULA.BiCopName(copula, False)[0])

    if copula in [1, 2]:
        return -1., 1.
    elif copula in [3, 13]:
        return 0., np.inf
    elif copula in [23, 33]:
        return -np.inf, 0.
    elif copula in [4, 14, 24, 34]:
        return 1., np.inf
    elif copula in [5]:
        return -np.inf, np.inf
    elif copula in [6, 26, 36]:
        return 1., np.inf
    else:
        raise NotImplementedError("Not implemented yet.")


def get_param2_interval(copula):
    """
    """
    assert isinstance(copula, (np.integer, str)), \
        TypeError("Input must be int or str")

    if copula in [2, 't']:
        return 1.e-6, np.inf
    else:
        raise NotImplementedError("Not implemented yet.")


def get_tau_interval(copula):
    assert isinstance(copula, (np.integer, str)), \
        TypeError("Input must be int or str. Not: ", type(copula))
    if isinstance(copula, str):
        copula = int(R_VINECOPULA.BiCopName(copula, False)[0])

    if copula in [1, 2]:
        return -0.99, 0.99
    elif copula in [3, 13, 4, 14, 5, 6, 16]:
        return 0.01, 0.98
    elif copula in [23, 24, 26, 33, 34, 36]:  # Rotated copulas
        return -0.99, -0.01
    else:
        raise NotImplementedError("Not implemented yet.")


class Conversion(object):
    """
    Static class to convert dependence parameters.
    
    Parameters
    ----------
    family : array,
        The family matrix.
    """    
    def __init__(self, family):
        self.family = family

    def to_copula_parameter(self, measure_param, dep_measure='kendall-tau'):
        """Convert the dependence_measure to the copula parameter.
        
        Parameters
        ----------
        
        
        Returns
        -------
        """
        if isinstance(measure_param, list):
            measure_param = np.asarray(measure_param)

        if isinstance(measure_param, np.ndarray):
            n_sample = measure_param.shape[0]
            assert n_sample == measure_param.size, "It must be a vector"
        elif isinstance(measure_param, float):
            n_sample = 1
            measure_param = np.asarray([measure_param])
        else:
            raise TypeError("Wrong type for measure_param: {0}".format(type(measure_param)))
        
        if dep_measure == "kendall-tau":
            r_kendall = numpy2ri(measure_param)
            copula_param = np.asarray(R_VINECOPULA.BiCopTau2Par(self._family, r_kendall))
        elif dep_measure == "pearson-rho":
            copula_param = self._copula.fromPearsonToParam(measure_param)
        else:
            raise ValueError("Unknow Dependence Measure")

        if copula_param.size == 1:
            return copula_param.item()
        else:
            return copula_param
    
    def to_kendall(self, params):
        """Convert the dependence_measure to the copula parameter.      
        
        Parameters
        ----------
        
        
        Returns
        -------
        """
        if isinstance(params, np.ndarray):
            r_params = numpy2ri(params)
        elif isinstance(params, float):
            r_params = params
        else:
            raise TypeError("Wrong type for params. Got: ", type(params))
        copula_param = np.asarray(R_VINECOPULA.BiCopPar2Tau(self._family, r_params))
        return copula_param

    def to_pearson(self, measure_param):
        """Convert the dependence_measure to the copula parameter.
                
        Parameters
        ----------
        
        
        Returns
        -------
        """
        return self._copula.fromParamToPearson(measure_param)
        
    @property
    def family(self):
        """The family matrix.
        """
        return self._family
        
    @family.setter
    def family(self, value):
        if isinstance(value, (int, float, np.integer)):
            np.testing.assert_equal(value, int(value))
            self._family = int(value)
            self._family_name = R_VINECOPULA.BiCopName(int(value), False)[0]            
        elif isinstance(value, str):
            self._family = int(R_VINECOPULA.BiCopName(value, False)[0])
            self._family_name = value
        else:
            raise TypeError("Unkow Type for family")

    class NormalCopula:
        """
        """
        @staticmethod
        def to_kendall(param):
            """
            From Pearson correlation parameter to Kendal dependence parameter.        
            Parameters
            ----------
            
            
            Returns
            -------
            """
            return 2. / np.pi * np.arcsin(param)

        @staticmethod
        def to_copula_parameter(tau):
            """
            From Kendal dependence parameter to Pearson correlation parameter.
                    
            Parameters
            ----------
            
            
            Returns
            -------
            """
            return np.sin(np.pi / 2. * tau)



import copy
import os

import numpy as np

from .conservative import ListDependenceResult
from .utils import get_pair_id, get_pairs_by_levels, get_possible_structures

GRIDS = ['lhs', 'rand', 'vertices']
LIB_PARAMS = ['iterative_save', 'iterative_load', 'input_names', 
              'output_names', 'keep_input_samples', 'load_input_samples',
              'use_grid', 'save_grid', 'grid_path', 'n_pairs_start']


def iterative_vine_minimize(estimate_object, n_input_sample=1000, n_dep_param_init=20, max_n_pairs=5, grid_type='lhs', 
                            q_func=np.var, n_add_pairs=1, n_remove_pairs=0, adapt_vine_structure=True, delta=0.1,
                            with_bootstrap=False, verbose=False, **kwargs):
    """Iteratively minimises the output quantity of interest.

    Parameters
    ----------


    Returns
    -------
    """
    quant_estimate = copy.copy(estimate_object)
    corr_dim = quant_estimate.corr_dim
    dim = quant_estimate.input_dim
    
    max_n_pairs = min(max_n_pairs, corr_dim)
    
    assert grid_type in GRIDS, "Unknow Grid type {0}".format(grid_type)
    assert 0 < max_n_pairs <= corr_dim, "Maximum number of pairs must be positive"
    assert 1 <= n_add_pairs <= corr_dim, "Must add at least one pair at each iteration"
    assert 0 <= n_remove_pairs < corr_dim, "This cannot be negative"
    assert callable(q_func), "Quantity function must be callable"
    
    if n_add_pairs == corr_dim:
        adapt_vine_structure = False
        print('The number of dimension is equal to the number of pairs to add')
        
    # Check if the given parameters are known
    for lib_param in kwargs:
        assert lib_param in LIB_PARAMS, "Unknow parameter %s" % (lib_param)
        
    # Iterative save of the results
    iterative_save = False
    if 'iterative_save' in kwargs:
        iterative_save = kwargs['iterative_save']
        if iterative_save is True:
            save_dir = './iterative_result'
        elif isinstance(iterative_save, str):
            save_dir = os.path.abspath(iterative_save)
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
        elif iterative_save is False:
            pass
        else:
            raise TypeError("Wrong type for iterative_save: {}".format(type(iterative_save)))
        
    # Iterative load of the results
    iterative_load = False
    if 'iterative_load' in kwargs:
        iterative_load = kwargs['iterative_load']
        if iterative_load is True:
            load_dir = './iterative_result'
        elif isinstance(iterative_load, str):
            load_dir = os.path.abspath(iterative_load)
            if not os.path.exists(load_dir):
                print("Directory %s does not exists" % (load_dir))
        elif iterative_load is False:
            pass
        else:
            raise TypeError("Wrong type for iterative_load: {0}".format(type(iterative_load)))
            
    input_names = []
    if 'input_names' in kwargs:
        input_names = kwargs['input_names']

    output_names = []
    if 'output_names' in kwargs:
        output_names = kwargs['output_names']
        
    keep_input_samples = True
    if 'keep_input_samples' in kwargs:
        keep_input_samples = kwargs['keep_input_samples']
        
    load_input_samples = True
    if 'load_input_samples' in kwargs:
        load_input_samples = kwargs['load_input_samples']
        
    use_grid = None
    if 'use_grid' in kwargs:
        use_grid = kwargs['use_grid']
        
    save_grid = None
    if 'save_grid' in kwargs:
        save_grid = kwargs['save_grid']
                
    n_pairs_start = 0
    if 'n_pairs_start' in kwargs:
        n_pairs_start = kwargs['n_pairs_start']

    # Only a loading execution
    if n_input_sample == 0:
        iterative_save = None
        
    # Initial configurations
    init_family = quant_estimate.families
    init_bounds_tau = quant_estimate.bounds_tau
    fixed_params = quant_estimate.fixed_params.copy()
    init_indep_pairs = quant_estimate._indep_pairs[:]
    init_fixed_pairs = quant_estimate._fixed_pairs[:]
    
    # New empty configurations
    families = np.zeros((dim, dim))
    bounds_tau = np.zeros((dim, dim))
    bounds_tau[:] = None

    # Selected pairs through iterations
    selected_pairs = []
    all_results = IterativeDependenceResults(dim)

    if callable(n_dep_param_init):
        n_param_iter = n_dep_param_init
    elif n_dep_param_init is None:
        n_param_iter = lambda x: None
    else:
        n_param_iter = lambda k: int(n_dep_param_init*(k+1)**2)

    n_dep_param = n_param_iter(0)
    
    # The pairs to do at each iterations
    indices = np.asarray(np.tril_indices(dim, k=-1)).T.tolist()
    
    # Remove fixed pairs from the list and add in the family matrix
    for pair in init_fixed_pairs:
        indices.remove(pair)
        families[pair[0], pair[1]] = init_family[pair[0], pair[1]]
        
    # Remove independent pairs
    for pair in init_indep_pairs:
        indices.remove(pair)
        
    ## Algorithm Loop
    cost = 0
    n_pairs = 1
    iteration = 0
    min_quant_iter = []
    stop_conditions = False
    
    while not stop_conditions:
        min_quantity = {}
        for i, j in indices:
            # Family matrix for this iteration
            tmp_families = families.copy()
            tmp_families[i, j] = init_family[i, j]
            tmp_bounds_tau = bounds_tau.copy()
            tmp_bounds_tau[i, j] = init_bounds_tau[i, j]
            tmp_bounds_tau[j, i] = init_bounds_tau[j, i]
            
            # Adapt the vine structure matrix
            if adapt_vine_structure:
                pairs_iter = init_indep_pairs + init_fixed_pairs + selected_pairs + [(i, j)]
                pairs_iter_id = [get_pair_id(dim, pair, with_plus=False) for pair in pairs_iter]
                pairs_by_levels = get_pairs_by_levels(dim, pairs_iter_id)
                quant_estimate.vine_structure = get_possible_structures(dim, pairs_by_levels)[0]
            
            # Family matrix is changed
            quant_estimate.families = tmp_families
            quant_estimate.fixed_params = fixed_params
            quant_estimate.bounds_tau = tmp_bounds_tau

            # Lets get the results for this family structure
            if n_input_sample > 0 and n_pairs >= n_pairs_start:
                results = quant_estimate.gridsearch(n_dep_param=n_dep_param,
                                                             n_input_sample=n_input_sample,
                                                             grid_type=grid_type,
                                                             keep_input_samples=keep_input_samples,
                                                             load_grid=use_grid,
                                                             save_grid=save_grid,
                                                             use_sto_func=True)
                results.q_func = q_func
            
            if iterative_save or iterative_load:
                cop_str = "_".join([str(l) for l in quant_estimate._family_list])
                vine_str = "_".join([str(l) for l in quant_estimate._vine_structure_list])
                filename = "%s/%s" % (load_dir, grid_type)
                if n_dep_param is None:
                    filename += "_K_None"
                else:
                    filename += "_K_%d" % (n_dep_param)
                filename += "_cop_%s_vine_%s.hdf5" % (cop_str, vine_str)

            if iterative_save and n_pairs >= n_pairs_start:
                results.to_hdf(filename, input_names, output_names, with_input_sample=keep_input_samples)

            if iterative_load :
                name, extension = os.path.splitext(filename)
                condition = os.path.exists(filename)
                k = 0
                while condition:
                    try:
                        load_result = ListDependenceResult.from_hdf(filename, with_input_sample=load_input_samples, q_func=q_func)
                        # TODO: create a function to check the configurations of two results
                        # TODO: is the testing necessary? If the saving worked, the loading should be ok.
                        np.testing.assert_equal(load_result.families, tmp_families, err_msg="Not good family")
                        np.testing.assert_equal(load_result.bounds_tau, tmp_bounds_tau, err_msg="Not good Bounds")
                        np.testing.assert_equal(load_result.vine_structure, quant_estimate.vine_structure, err_msg="Not good structure")
                        condition = False
                    except AssertionError:
                        filename = '%s_num_%d%s' % (name, k, extension)
                        condition = os.path.exists(filename)
                        k += 1
                # Replace the actual results with the loaded results (this results + all the previous saved ones)
                results = load_result
                
            # How much does it costs
            cost += results.n_evals

            # Save the minimum
            if not with_bootstrap:
                min_quantity[i, j] = results.min_quantity
            else:
                assert isinstance(with_bootstrap, int), "Must be a number"
                n_bootstrap = with_bootstrap
                results.compute_bootstraps(n_bootstrap)
                print(results.bootstrap_samples.mean(axis=1))
                min_quantity[i, j] = results[results.bootstrap_samples.mean(axis=1).argmin()]

            if verbose:
                print('n={}, K={}. Worst quantile of {} at {}'.format(results.n_input_sample, n_dep_param, selected_pairs + [(i, j)], min_quantity[i, j]))
                if input_names:
                    pair_names = [ "%s-%s" % (input_names[k1], input_names[k2]) for k1, k2 in selected_pairs + [(i, j)]]
                    print("The variables are: " + " ".join(pair_names))


            # Store the result
            all_results[iteration, i, j] = results

        # Get the min from the iterations
        sorted_quantities = sorted(min_quantity.items(), key=lambda x: x[1])
        
        # Delay of the first iteration
        if iteration == 0:
            delta_q_init = abs(sorted_quantities[0][1] - sorted_quantities[-1][1])

        min_quant_iter.append(sorted_quantities[0][1])
        
        if (n_remove_pairs > 0) and (n_remove_pairs < len(sorted_quantities)-1):
            # The pairs to remove
            for pair in sorted_quantities[-n_remove_pairs:]:
                indices.remove(list(pair[0]))

        selected_pair = sorted_quantities[0][0]
        # Selected pairs to add 
        for pair in sorted_quantities[:n_add_pairs]:
            i, j = pair[0][0], pair[0][1]
            families[i, j] = init_family[i, j]
            bounds_tau[i, j] = init_bounds_tau[i, j]
            bounds_tau[j, i] = init_bounds_tau[j, i]
            indices.remove(list(pair[0]))
            selected_pairs.append(pair[0])
            
        all_results.selected_pairs.append(selected_pairs)
        if True:
            k1, k2 = selected_pair
            tmp = '\nIteration {0}: selected pair: {1}'.format(iteration+1, selected_pair)
            if input_names:
                tmp += " (" + "-".join(input_names[k] for k in selected_pair) + ")"
            print(tmp)
            print('Total number of evaluations = %d. Minimum quantity at %.2f.\n' % (cost, min_quantity[selected_pair]))

        # Stop conditions
        if n_pairs >= max_n_pairs:
            stop_conditions = True
            print('Max number of pairs reached')
            
        if iteration > 0:
            delta_q = -(min_quant_iter[-1] - min_quant_iter[-2])
            if delta_q <= delta*delta_q_init:
                stop_conditions = True
                print('Minimum_variation not fulfiled: %.2f <= %0.2f' % (delta_q, delta*delta_q_init))
        n_pairs += n_add_pairs
        if n_dep_param is not None:
            n_dep_param = n_param_iter(iteration+1)

        if not stop_conditions:
            all_results.new_iteration()

        all_results.n_evals = cost

        iteration += 1
    return all_results


class IterativeDependenceResults(object):
    """

    """
    def __init__(self, dim):

        self.iteration = 0
        n_pairs = int(dim * (dim-1) / 2)
        self.results = [[]]
        tmp = np.zeros((dim, dim), dtype=object)
        tmp[:] == None
        self.results[self.iteration] = tmp
        self.selected_pairs = [[]]
        
        self.dim = dim
        self.n_pairs = n_pairs
        self.n_evals = 0

    def new_iteration(self):
        """
        """
        self.iteration += 1
        tmp = np.zeros((self.dim, self.dim), dtype=object)
        tmp[:] = None
        self.results.append(tmp)

    def __getitem__(self, item):
        """
        """
        iteration, i, j = item
        return self.results[iteration][i, j]

    def __setitem__(self, item, result):
        """
        """
        iteration, i, j = item
        self.results[iteration][i, j] = result

    def min_quantities(self, iteration):
        """
        """
        results = self.results[iteration]
        dim = self.dim
        min_quantities = np.zeros((dim, dim), dtype=np.float)
        for i in range(1, dim):
            for j in range(i):
                if results[i, j] is not None:
                    min_quantities[i, j] = results[i, j].min_quantity

        return min_quantities

    def min_results(self, iteration):
        """
        """
        results = self.results[iteration]
        dim = self.dim
        min_results = np.zeros((dim, dim), dtype=object)
        for i in range(1, dim):
            for j in range(i):
                if results[i, j] is not None:
                    min_results[i, j] = results[i, j].min_result

        return min_results

    def min_quantity(self, iteration):
        """
        """
        min_quantities = self.min_quantities(iteration)
        min_quantity = min_quantities.min()
        return min_quantity


    def min_result(self, iteration):
        """
        """
        min_quantities = self.min_quantities(iteration)
        id_min = min_quantities.argmin()
        min_result = self.min_results(iteration).item(id_min)
        return min_result



import copy
import operator
import os
from itertools import permutations, product

import numpy as np
import openturns as ot
from pyDOE import lhs
from sklearn.utils import check_random_state
from sklearn.utils.fixes import sp_version

from skopt.space import Space as sk_Space

OPERATORS = {">=": operator.ge,
             ">": operator.gt,
             "==": operator.eq}

MAX_N_PAIR_VERTICES = 13


def get_grid_sample(dimensions, n_sample, grid_type):
    """Sample inside a fixed design space.

    Parameters
    ----------
    dimensions : array,
        The bounds of the space for each dimensions.
    n_sample: int,
        The number of observations inside the space.
    grid_type: str,
        The type of sampling.

    Returns
    -------
    sample : array,
        The sample from the given space and the given grid.
    """
    # We create the grid
    space = Space(dimensions)
    sample = space.rvs(n_sample, sampling=grid_type)
    return sample


class Space(sk_Space):
    """
    """

    def rvs(self, n_samples=1, sampling='rand',
            lhs_sampling_criterion='centermaximin', random_state=None):
        """Draw random samples.

        The samples are in the original space. They need to be transformed
        before being passed to a model or minimizer by `space.transform()`.

        Parameters
        ----------
        n_samples : int or None, optional (default=1)
            Number of samples to be drawn from the space. If None and 
            sampling is 'vertices', then all the space vertices are taken.

        sampling : str, optional (default='rand')
            The sampling strategy, which can be:
            - 'rand' : Random sampling
            - 'lhs' : Latin Hypercube Sampling is done
            - 'vertices' : Sampling over the vertices of the space.

        lhs_sampling_criterion : str, optional (default='centermaximin')
            The sampling criterion for the LHS sampling.

        random_state : int, RandomState or None, optional (default=None)
            Set random state to something other than None for reproducible
            results.

        Returns
        -------
        points : list of lists, shape=(n_points, n_dims)
           Points sampled from the space.
        """
        rng = check_random_state(random_state)
        assert isinstance(sampling, str), \
            TypeError("sampling must be a string")

        if sampling == 'rand':
            # Random sampling
            columns = []
            for dim in self.dimensions:
                if sp_version < (0, 16):
                    columns.append(dim.rvs(n_samples=n_samples))
                else:
                    columns.append(
                        dim.rvs(n_samples=n_samples, random_state=rng))
            # Transpose
            rows = []
            for i in range(n_samples):
                r = []
                for j in range(self.n_dims):
                    r.append(columns[j][i])
                rows.append(r)
        elif sampling == 'lhs':
            # LHS sampling
            sample = lhs(self.n_dims, samples=n_samples,
                         criterion=lhs_sampling_criterion)
            tmp = np.zeros((n_samples, self.n_dims))
            # Assert the bounds
            for k, dim in enumerate(self.dimensions):
                tmp[:, k] = sample[:, k]*(dim.high - dim.low) + dim.low
            rows = tmp.tolist()
        elif sampling == 'vertices':
            # Sample on the vertices of the space.
            n_pair = len(self.dimensions)

            if n_pair > MAX_N_PAIR_VERTICES:
                if n_samples is None:
                    raise MemoryError(
                        'Too much pairs to create a vertices grid.')

                # Iterative add instead of sampling in a big bounds matrix
                sample = []
                for i in range(n_samples):
                    obs = np.random.choice([-1., 1., 0.], size=n_pair).tolist()
                    if obs not in sample:
                        sample.append(obs)

                sample = np.asarray(sample)
            else:
                bounds = list(product([-1., 1., 0.], repeat=n_pair))
                if n_samples is None or n_pair == 1:
                    bounds.remove((0.,)*n_pair)  # remove indepencence
                bounds = np.asarray(bounds)
                n_bounds = len(bounds)

                if (n_samples is None) or (n_samples > n_bounds):
                    # We take all the vertices
                    n_samples = n_bounds
                    sample = bounds
                else:
                    id_taken = np.random.choice(
                        n_bounds, size=n_samples, replace=False)
                    sample = bounds[sorted(id_taken), :]
            # Assert the bounds
            for p in range(n_pair):
                sample_p = sample[:, p]
                sample_p[sample_p == -1.] = self.dimensions[p].low
                sample_p[sample_p == 1.] = self.dimensions[p].high

            rows = sample.tolist()
        elif sampling == 'fixed':
            raise NotImplementedError("Maybe I'll do it...")
        else:
            raise NameError("Sampling type does not exist.")
        return rows


def list_to_matrix(values, dim):
    """Transform a list of values in a lower triangular matrix.

    Parameters
    ----------
    param : list, array
        The list of values
    dim : int,
        The shape of the matrix

    Returns
    -------
    matrix : array
        The lower triangular matrix
    """
    matrix = np.zeros((dim, dim))
    k = 0
    for i in range(1, dim):
        for j in range(i):
            matrix[i, j] = values[k]
            k += 1

    return matrix


def matrix_to_list(matrix, return_ids=False, return_coord=False, op_char='>'):
    """Convert a lower triangular martix into a list of its values.

    Parameters
    ----------
    matrix : array
        A square lower triangular matrix.
    return_ids : bool, optional (default=False)
        If true, the index of the list are returns.
    return_coord : bool, optional (default=False)
        If true, the coordinates in the matrix are returns.
    op_char : str, optional (default='>')
        If '>', only the positive values of the matrix are taken into account.

    Returns
    -------
    values : list
        The list of values in the matrix. If only_positive is True, then only the positive values
        are returns.
    """
    op_func = OPERATORS[op_char]
    values = []
    ids = []
    coord = []
    dim = matrix.shape[0]
    k = 0
    for i in range(1, dim):
        for j in range(i):
            if op_func(matrix[i, j], 0):
                values.append(matrix[i, j])
                ids.append(k)
                coord.append([i, j])
            k += 1

    if return_ids and return_coord:
        return values, ids, coord
    elif return_ids:
        return values, ids
    elif return_coord:
        return values, coord
    else:
        return values


def bootstrap(data, num_samples, statistic):
    """Returns bootstrap estimate of 100.0*(1-alpha) CI for statistic.

    Inspired from: http://people.duke.edu/~ccc14/pcfb/analysis.html"""
    n = len(data)
    idx = np.random.randint(0, n, (num_samples, n))
    samples = data[idx]
    stat = np.sort(statistic(samples, axis=1))
    return stat


def to_kendalls(converters, params):
    """Convert the copula parameters to kendall's tau.

    Parameters
    ----------
    converter s: list of VineCopula converter
        The converters from the copula parameter to the kendall tau for the given families.
    params : list or array
        The parameters of each copula converter.

    Returns
    -------
    kendalls : list
        The kendalls tau of the given parameters of each copula
    """
    if isinstance(params, list):
        params = np.asarray(params)
    elif isinstance(params, float):
        params = np.asarray([params])

    n_params, n_pairs = params.shape
    kendalls = np.zeros(params.shape)
    for k in range(n_pairs):
        kendalls[:, k] = converters[k].to_kendall(params[:, k])

    # If there is only one parameter, no need to return the list
    if kendalls.size == 1:
        kendalls = kendalls.item()
    return kendalls


def to_copula_params(converters, kendalls):
    """Convert the kendall's tau to the copula parameters.

    Parameters
    ----------
    converters : list of VineCopula converters
        The converters from the kendall tau to the copula parameter of the given families.
    kendalls : list or array
        The kendall's tau values of each pairs.
    Returns
    -------
    params : array
        The copula parameters.
    """
    if isinstance(kendalls, list):
        kendalls = np.asarray(kendalls)
    elif isinstance(kendalls, float):
        kendalls = np.asarray([kendalls])

    n_params, n_pairs = kendalls.shape
    params = np.zeros(kendalls.shape)
    for k in range(n_pairs):
        params[:, k] = converters[k].to_copula_parameter(
            kendalls[:, k], dep_measure='kendall-tau')

    # If there is only one parameter, no need to return the list
    if params.size == 1:
        params = params.item()
    return params


def margins_to_dict(margins):
    """Convert the margin's informations into a dictionary.

    Parameters
    ----------
    margins : the list of OpenTurns distributions
        The marginal distributions of the input variables.

    Returns
    -------
    margin_dict : dict
        The dictionary with the information of each marginal.
    """
    margin_dict = {}
    for i, marginal in enumerate(margins):
        margin_dict[i] = {}
        name = marginal.getName()
        params = list(marginal.getParameter())
        if name == 'TruncatedDistribution':
            margin_dict[i]['Type'] = 'Truncated'
            in_marginal = marginal.getDistribution()
            margin_dict[i]['Truncated Parameters'] = params
            name = in_marginal.getName()
            params = list(in_marginal.getParameter())
        else:
            margin_dict[i]['Type'] = 'Standard'

        margin_dict[i]['Marginal Family'] = name
        margin_dict[i]['Marginal Parameters'] = params
    return margin_dict


def dict_to_margins(margin_dict):
    """Convert a dictionary with margins informations into a list of distributions.

    Parameters
    ----------
    margin_dict : dict
        A dictionary of information on the margins

    Returns
    -------
    margins
    """
    margins = []
    for i in sorted(margin_dict.keys()):
        marginal = getattr(ot, margin_dict[i]['Marginal Family'])(
            *margin_dict[i]['Marginal Parameters'])
        if margin_dict[i]['Type'] == 'TruncatedDistribution':
            params = margin_dict[i]['Bounds']
            marginal = ot.TruncatedDistribution(marginal, *params)
        margins.append(marginal)

    return margins


def save_dependence_grid(dirname, kendalls, bounds_tau, grid_type):
    """Save a grid of kendall's into a csv ifile.

    The grid is always saved in Kendall's Tau measures.

    Parameters
    ----------
    dirname : str
        The directory path.

    kendalls : list or array
        The kendall's tau of each pair.

    bounds_tau : list or array
        The bounds on the kendall's tau.

    grid_type : str
        The ype of grid.

    Returns
    -------
    grid_filename : str
        The grid filename
    """
    kendalls = np.asarray(kendalls)
    n_param, n_pairs = kendalls.shape

    # The sample variable to save
    sample = np.zeros((n_param, n_pairs))
    for k in range(n_pairs):
        tau_min, tau_max = bounds_tau[k]
        sample[:, k] = (kendalls[:, k] - tau_min) / (tau_max - tau_min)

    k = 0
    do_save = True
    name = '%s_p_%d_n_%d_%d.csv' % (grid_type, n_pairs, n_param, k)

    grid_filename = os.path.join(dirname, name)
    # If this file already exists
    while os.path.exists(grid_filename):
        existing_sample = np.loadtxt(grid_filename).reshape(n_param, -1)
        # We check if the build sample and the existing one are equivalents
        if np.allclose(existing_sample, sample):
            do_save = False
            print('The DOE already exist in %s' % (name))
            break
        k += 1
        name = '%s_p_%d_n_%d_%d.csv' % (grid_type, n_pairs, n_param, k)
        grid_filename = os.path.join(dirname, name)

    # It is saved
    if do_save:
        np.savetxt(grid_filename, sample)
        print("Grid saved at %s" % (grid_filename))

    return grid_filename


def load_dependence_grid(dirname, n_pairs, n_params, bounds_tau, grid_type, use_grid=None):
    """Load a grid of parameters

    Parameters
    ----------
    dirname : str
        The directory path.

    n_params : int
        The grid dimension (the number of dependent pairs).

    n_params : int
        The grid size of the sample.

    bounds_tau : list or array
        The bounds on the kendall's tau.

    grid_type : str
        The ype of grid.

    use_grid : int, str or None, optional (default=None)
        If a particular grid should be used.

    Returns
    -------
    kendalls : array
        The kendall's tau of each dependent pairs.
    filename : str
        The name of the loaded grid.
    """
    if isinstance(use_grid, str):
        filename = use_grid
        name = os.path.basename(filename)
    elif isinstance(use_grid, (int, bool)):
        k = int(use_grid)
        name = '%s_p_%d_n_%d_%d.csv' % (grid_type, n_pairs, n_params, k)
        filename = os.path.join(dirname, name)
    else:
        raise AttributeError('Unknow use_grid')

    assert os.path.exists(filename), 'Grid file %s does not exists' % name
    print('loading file %s' % name)
    sample = np.loadtxt(filename).reshape(n_params, n_pairs)
    assert n_params == sample.shape[0], 'Wrong grid size'
    assert n_pairs == sample.shape[1], 'Wrong dimension'

    kendalls = np.zeros((n_params, n_pairs))
    for k in range(n_pairs):
        tau_min, tau_max = bounds_tau[k]
        kendalls[:, k] = sample[:, k]*(tau_max - tau_min) + tau_min

    return kendalls, filename


def quantile_func(alpha):
    """To associate an alpha to an empirical quantile function.

    Parameters
    ----------
    alpha : float
        The probability of the target quantile. The value must be between 0 and 1.

    Returns
    -------
    q_func : callable
        The quantile function.

    """
    def q_func(x, axis=1):
        return np.percentile(x, alpha*100., axis=axis)
    return q_func


def proba_func(threshold):
    """To associate an alpha to an empirical distribution function.

    Parameters
    ----------
    threshold : float
        The threshold of the target probability.

    Returns
    -------
    p_func : callable
        The probability function.

    """
    def p_func(x, axis=1):
        return (x >= threshold).mean(axis=axis)
    return p_func


def asymptotic_error_quantile(n, q_density, q_alpha):
    """
    """
    return np.sqrt(q_alpha * (1. - q_alpha) / (n * q_density**2))


def asymptotic_error_proba(n, proba):
    """
    """
    return np.sqrt(proba * (1. - proba) / n)


def get_pairs_by_levels(dim, forced_pairs_ids, verbose=False):
    """Split a list of pairs in levels.

    Parameters
    ----------
    dim : int,
        The number of variables.
    forced_pairs_ids : list of int,
        The list of pair ids.
    verbose : bool,
        If True, informations about the procedure are printed.

    Returns
    -------
    pairs_by_levels : list of list,
        The list of listed pairs.
    """
    n_pairs = int(dim*(dim-1)/2)
    n_forced_pairs = len(forced_pairs_ids)
    assert len(np.unique(forced_pairs_ids)) == n_forced_pairs, \
        "The list should contain unique values"
    assert n_forced_pairs <= n_pairs, "Too many pairs: %d > %d" % (
        n_forced_pairs, n_pairs)
    assert max(forced_pairs_ids) < n_pairs, "Wrong pair id: %d" % (
        max(forced_pairs_ids))

    forced_pairs = get_pairs(dim, forced_pairs_ids, with_plus=True)

    remaining_pairs_ids = list(range(0, n_pairs))
    for pair_id in forced_pairs_ids:
        remaining_pairs_ids.remove(pair_id)
    remaining_pairs = get_pairs(dim, remaining_pairs_ids, with_plus=True)

    if verbose:
        print('Vine dimension: %d' % dim)
        print('Conditioning information:')

    k0 = 0
    pairs_by_levels = []

    # For each level and the number of slot per level
    for lvl, n_slot_lvl in enumerate(range(dim-1, 0, -1)):
        k1 = min(n_forced_pairs, k0+n_slot_lvl)
        if forced_pairs[k0:k0+n_slot_lvl]:
            pairs_by_levels.append(forced_pairs[k0:k0+n_slot_lvl])

        if verbose:
            print("\t%d pairs with %d conditionned variables" %
                  (n_slot_lvl, lvl))
            print("Pairs: {0}".format(forced_pairs[k0:k0+n_slot_lvl]))
        k0 = k1

    if verbose:
        print("Concerned pairs: {0}".format(forced_pairs))
        print("Remaining pairs: {0}".format(remaining_pairs))

    # This part checks if the pairs in a same level respect the condition.
    # If not, a permutation of between levels is made.
    good_pairs_by_levels = copy.deepcopy(pairs_by_levels)  # copy
    n_moving_pairs = [1]*(dim-1)
    loops = [0]*(dim-1)
    move_up = [1]*(dim-1)

    while not check_conditions(pairs_by_levels, dim):
        pairs_by_levels = copy.deepcopy(good_pairs_by_levels)
        n_levels = len(pairs_by_levels)
        lvl = 0
        while lvl < n_levels:
            pairs_lvl = pairs_by_levels[lvl]
            if verbose:
                print("Level:", lvl)
                print("Before:", pairs_lvl)
            if not check_condition(pairs_lvl, lvl, dim):
                # If it's the last level
                idx = loops[lvl]
                if lvl == n_levels - 1:
                    pairs_lvl_up = []
                    # A new level is created
                    n_moving_pair = n_moving_pairs[lvl]

                    # The last pair is taken back from the level
                    moving_pairs = []
                    for k in range(n_moving_pair-1):
                        moving_pairs.append(pairs_lvl.pop())
                    moving_pairs.append(pairs_lvl.pop(-idx-1))
                    # It is added in the next level (which is created)
                    pairs_by_levels.append(moving_pairs)
                    n_levels += 1
                else:
                    # The 1st pair of the next level is switch with the last
                    # of the previous
                    pairs_lvl_up = pairs_by_levels[lvl+1]
                    n_moving_pair = n_moving_pairs[lvl]
                    for k in range(n_moving_pair-1):
                        pairs_lvl_up.insert(0, pairs_lvl.pop())
                    pairs_lvl_up.insert(0, pairs_lvl.pop(-idx-1))

                    if move_up[lvl] > 0:
                        id_up = move_up[lvl]
                        # The pair is moved
                        moving_pair_up = pairs_lvl_up.pop(id_up)
                        pairs_lvl.append(moving_pair_up)
                if verbose:
                    print("After:", pairs_lvl)
                    print("n_moving:", n_moving_pairs[lvl])
                    print("idx:", loops[lvl])
                    print("move_up:", move_up[lvl])
                    print("Is OK?", check_condition(pairs_lvl, lvl, dim))

                if move_up[lvl] == 0:
                    loops[lvl] += 1
                else:
                    move_up[lvl] += 1

                if move_up[lvl] >= len(pairs_lvl_up):
                    move_up[lvl] = 0

                if loops[lvl] >= len(pairs_lvl):
                    loops[lvl] = 0
                    n_moving_pairs[lvl] += 1

                # This level does not work. Lets reset the options of the others
                for lvl_up in range(lvl+1, n_levels):
                    move_up[lvl_up] = 1
                    n_moving_pairs[lvl_up] = 1
                    loops[lvl_up] = 0

                # If still not good, we reloop
                if not check_condition(pairs_lvl, lvl, dim):
                    break
                else:
                    good_pairs_by_levels[lvl] = pairs_lvl

            lvl += 1

    return pairs_by_levels


def check_pairs_number_level(pairs_lvl, lvl, dim):
    """
    """
    if len(pairs_lvl) <= dim-1-lvl:
        return True
    else:
        return False


def check_condition(pairs_lvl, lvl, dim):
    """
    """

    return check_node_loop(pairs_lvl) and check_pairs_number_level(pairs_lvl, lvl, dim)


def check_conditions(pairs_by_levels, dim):
    """
    """
    for lvl, pairs_lvl in enumerate(pairs_by_levels):
        if not check_condition(pairs_lvl, lvl, dim):
            return False
    return True


def add_pair(structure, pair, index, lvl):
    """Adds a pair in a Vine structure in a certain place and for a specific conditionement.

    Parameters
    ----------
    structure : array,
        The R-vine structure array.
    pair : list,
        A pair of variables that we want to add.
    index : int, 
        The position of the pair of variable.
    lvl : int,
        The tree level.

    Returns
    -------
    structure : array,
        The updated R-vine structure.
    """
    dim = structure.shape[0]
    if lvl == 0:  # If it's the unconditiononal variables
        assert structure[index, index] == 0, \
            "There is already a variable at [%d, %d]" % (index, index)
        assert structure[dim-1, index] == 0, \
            "There is already a variable at [%d, %d]" % (dim-1, index)
        structure[index, index] = pair[0]
        structure[dim-1, index] = pair[1]
    else:
        assert structure[index, index] == pair[0], \
            "Diagonal element of the structure should be the same as the first variable of the pair: %d != %d" % (
                structure[index, index], pair[0])
        assert structure[dim-1-lvl, index] == 0, \
            "There is already a variable at [%d, %d]" % (dim-1-lvl, index)
        structure[dim-1-lvl, index] = pair[1]
    return structure


def add_pairs(structure, pairs, lvl, verbose=False):
    """Adds pairs in a structure for a selected level of conditionement.

    Parameters
    ----------
    structure : array,
        The R-vine structure array.
    pairs : list,
        The list of pairs of variables.
    lvl : int,
        The tree level for which we want the variables to be in.
    verbose : bool,
        If True, informations about the procedure are printed.
    """
    dim = structure.shape[0]
    n_pairs = len(pairs)

    assert n_pairs < dim - \
        lvl, "Not enough place to fill the pairs : %d > %d" % (
            n_pairs, dim-lvl)

    # Let's find a way to place these pairs inside the structure
    n_slots = dim - 1 - lvl
    possibilities = list(permutations(range(n_slots), r=n_pairs))
    success = False
    for possibility in possibilities:
        try:
            # Add the pair in the possible order
            structure_modif = np.copy(structure)
            for i in range(n_pairs):
                structure_modif = add_pair(
                    structure_modif, pairs[i], possibility[i], lvl)

            if check_redundancy(structure_modif):
                success = True
                break
        except AssertionError:
            pass

    if not success and verbose:
        print('Did not succeded to fill the structure with the given pairs')

    # If it's the 1st level, the last row of last column must be filled
    if (lvl == 0) and (n_pairs == dim-1):
        structure_modif[dim-1, dim -
                        1] = np.setdiff1d(range(dim+1), np.diag(structure_modif))[0]
    return structure_modif


def check_redundancy(structure):
    """Check if there is no redundancy of the diagonal variables.
    """
    dim = structure.shape[0]
    diag = np.diag(structure)
    for i in range(dim-1):
        # Check if it does not appears later in the matrix
        if diag[i] != 0.:
            if diag[i] in structure[:, i+1:]:
                return False
    return True


def rotate_pairs(init_pairs, rotations):
    """Rotate the pairs according to some rotations.
    """
    n_pairs = len(init_pairs)
    assert len(rotations) == n_pairs, \
        "The number of rotations is different to the number of pairs %d != %d" % (
            len(rotations), n_pairs)
    assert not np.setdiff1d(np.unique(rotations), [1, -1]), \
        "The rotations list should only be composed of -1 and 1."
    pairs = []
    for i in range(n_pairs):
        if rotations[i] == -1:
            pairs.append(list(reversed(init_pairs[i])))
        else:
            pairs.append(init_pairs[i])
    return pairs


def get_possible_structures(dim, pairs_by_levels, verbose=False):
    """Get the possible R-vine array for a given list of pairs by levels.
    """
    # For each levels
    good_structures = []
    for lvl, pairs_lvl in enumerate(pairs_by_levels):
        n_pairs_lvl = len(pairs_lvl)  # Number of pairs in the level

        # The possible combinations
        combinations = list(product([1, -1], repeat=n_pairs_lvl))
        if verbose:
            print("Number of combinations:", len(combinations))
        # Now lets get the possible pair combinations for this level
        for k, comb_k in enumerate(combinations):

            # Rotate the pair to the actual combination
            pairs_k = rotate_pairs(pairs_lvl, comb_k)
            if lvl == 0:
                # Create the associated vine structure
                structure = np.zeros((dim, dim), dtype=int)
                structure = add_pairs(structure, pairs_k, lvl, verbose=verbose)
                if check_redundancy(structure):
                    #                    structure[dim-2, dim-3] = np.setdiff1d(range(1, dim+1), np.diag(structure)[:dim-2].tolist() + [structure[dim-1, dim-3]])[0]
                    good_structures.append(structure)
            else:
                for structure in good_structures:
                    try:
                        new_structure = add_pairs(
                            structure, pairs_k, lvl, verbose=verbose)
                    except:
                        print(
                            "Can't add the pairs {0} in the current structure...".format(pairs_k))
                    if check_redundancy(new_structure):
                        structure = new_structure

    remain_structures = []
    for structure in good_structures:
        tmp = fill_structure(structure)
        if is_vine_structure(tmp):
            structure = tmp
            remain_structures.append(tmp)
            if verbose:
                print('good:\n{0}'.format(structure))

    return remain_structures


def check_structure_shape(structure):
    """Check if the structure shape is correct.
    """
    assert structure.shape[0] == structure.shape[1], "Structure matrix should be squared"
    assert np.triu(structure, k=1).sum(
    ) == 0, "Matrix should be lower triangular"


def check_natural_order(structure):
    """Check if a parent node is included in a child node.
    """
    d = structure.shape[0]
    for i in range(d-1):
        i = 1
        for j in range(i+1):
            parent = [[structure[j, j], structure[i+1, j]],
                      [structure[i+2:d, j].tolist()]]
            col = structure[:, j]
            parent_elements = col[np.setdiff1d(
                np.arange(j, d), range(j+1, i+1))]

            i_c = i + 1
            if len(parent_elements) > 2:
                n_child = 0
                for j_c in range(i_c+1):
                    possible_child = [
                        [structure[j_c, j_c], structure[i_c+1, j_c]], [structure[i_c+2:d, j_c].tolist()]]
                    col = structure[:, j_c]
                    possible_child_elements = col[np.setdiff1d(
                        np.arange(j_c, d), range(j_c+1, i_c+1))]
                    if len(np.intersect1d(possible_child_elements, parent_elements)) == d-i-1:
                        n_child += 1
                if n_child < 2:
                    return False

    return True


def is_vine_structure(matrix):
    """Check if the given matrix is a Vine Structure matrix
    """
    dim = matrix.shape[0]
    diag = np.diag(matrix)
    check_structure_shape(matrix)
    assert matrix.max() == dim, "Maximum should be the dimension: %d != %d" % (
        matrix.max(), dim)
    assert matrix.min() == 0, "Minimum should be 0: %d != %d" % (matrix.min(), dim)
    assert len(np.unique(diag)) == dim, "Element should be uniques on the diagonal: %d != %d" % (
        len(np.unique(diag)), dim)
    for i in range(dim):
        column_i = matrix[i:, i]
        assert len(np.unique(column_i)) == dim - \
            i, "Element should be unique for each column: %d != %d" % (
                len(np.unique(column_i)), dim - i)
        for node in diag[:i]:
            assert node not in column_i, "Previous main node should not exist after"

    if check_natural_order(matrix):
        return True
    else:
        return False


def check_node_loop(pairs):
    """Check if not too many variables are connected in a single tree.
    For 3 pairs of variables, it should have at least 4 variables involved.

    Parameters
    ----------
    pairs : tuple or list,
        The pair of variable.

    Returns
    -------
    """
    for n_l in range(3, len(pairs)+1):
        for perm_pairs in list(permutations(pairs, r=n_l)):
            if len(np.unique(perm_pairs)) <= n_l:
                return False
    return True


def fill_structure(structure):
    """Fill the structure with the remaining variables
    """
#    print structure
    dim = structure.shape[0]
#    structure[dim-2, dim-3] = np.setdiff1d(range(1, dim+1), np.diag(structure)[:dim-2].tolist() + [structure[dim-1, dim-3]])[0]

    diag = np.unique(np.diag(structure)).tolist()
    if 0 in diag:
        diag.remove(0)
    remaining_vals = np.setdiff1d(range(1, dim+1), diag)

    n_remaining_vals = len(remaining_vals)

    diag_possibility = list(permutations(remaining_vals, n_remaining_vals))[0]

    for k, i in enumerate(range(dim - n_remaining_vals, dim)):
        structure[i, i] = diag_possibility[k]

    structure[dim-1, dim-2] = structure[dim-1, dim-1]

    for i in range(dim - n_remaining_vals, dim-2):
        structure[dim-1, i] = structure[i+1, i+1]

    lvl = 1
    for j in range(dim-2, 0, -1):
        for i in range(j):
            col_i = structure[:, i]
            tmp = col_i[np.setdiff1d(np.arange(i, dim), range(j+1, i+1))]
            var_col_i = tmp[tmp != 0]
            for i_c in range(i+1, j+1):
                col_ic = structure[:, i_c]
                tmp = col_ic[np.setdiff1d(
                    np.arange(i_c, dim), range(j+1, i_c+1))]
                var_col_i_c = tmp[tmp != 0]
                intersection = np.intersect1d(var_col_i, var_col_i_c)
                if len(intersection) == lvl:
                    tt = var_col_i_c.tolist()
                    for inter in intersection:
                        tt.remove(inter)
                    structure[j, i] = tt[0]
                    break
        lvl += 1

    prev_ind = []
    for i in range(dim):
        values_i = structure[i:, i]
        used_values_i = values_i[values_i != 0].tolist() + prev_ind
        remaining_i = list(range(1, dim + 1))
        for var in used_values_i:
            if (var in remaining_i):
                remaining_i.remove(var)
        values_i[values_i == 0] = remaining_i
        prev_ind.append(values_i[0])
    return structure


def get_pair_id(dim, pair, with_plus=True):
    """ Get the pair id from a given index.    

    Parameters
    ----------
    dim : int,
        The dimension.
    pair : tuple of int,
        The pair of variable.
    with_plus : bool,
        If True, the variable numbers start at 1 instead of 0.

    Return
    ------
    pair_id : int,
        The pair index.
    """
    pairs = np.asarray(np.tril_indices(dim, k=-1)).T
    if with_plus:
        pair[0] += 1
        pair[1] += 1

    pair_id = np.where((pairs == pair).all(axis=1))[0][0]
    return pair_id


def get_pair(dim, index, with_plus=True):
    """ Get the pair of variables from a given index.

    Parameters
    ----------
    dim : int,
        The dimension.
    index : int,
        The ID of the pair.
    with_plus : bool,
        If True, the variable numbers start at 1 instead of 0.

    Return
    ------
    pair : tuple of int,
        The pair of variables.
    """
    pairs = np.asarray(np.tril_indices(dim, k=-1)).T
    pair = pairs[index]
    if with_plus:
        pair += 1
    pair = tuple(pair)

    return pair


def get_pairs(dim, pair_ids, with_plus=True):
    """Get the associated pairs of variables from given IDs.

    Parameters
    ----------
    dim : int,
        The dimension.
    pair_ids : list of int,
        The IDs of the pairs.
    with_plus : bool,
        If True, the variable numbers start at 1 instead of 0.

    Return
    ------
    pairs : list of tupple,
        The list of pairs.    
    """
    pairs = []
    for pair_id in pair_ids:
        pairs.append(get_pair(dim, pair_id, with_plus))

    pairs = list(map(tuple, pairs))
    return pairs



import numpy as np
import rpy2.rinterface as ri
from rpy2.robjects.packages import importr
from rpy2.robjects.numpy2ri import numpy2ri

VINECOPULA = importr('VineCopula')


def check_matrix(value):
    assert isinstance(value, np.ndarray), \
        TypeError('Variable must be a numpy array.')
    assert value.ndim == 2, \
        AttributeError('Matrix should be of dimension 2.')
    assert value.shape[0] == value.shape[1], \
        AttributeError('Matrix should be squared.')


def check_family(matrix):
    d = matrix.shape[0]
    for i in range(d):
        for j in range(i):
            if isinstance(matrix[i, j], str):
                matrix[i, j] = int(
                    VINECOPULA.BiCopName(matrix[i, j], False)[0])
            elif isinstance(matrix[i, j], np.integer):
                pass
            else:
                raise ValueError("Uncorrect Family matrix")

    return matrix


class VineCopula(object):
    """Vine Copula Class."""

    def __init__(self, structure, family, param1, param2=None):
        self.structure = structure
        self.family = family
        self.param1 = param1
        self.param2 = param2
        ri.initr()
        self.build_vine()

    @property
    def structure(self):
        return self._structure

    @structure.setter
    def structure(self, value):
        check_matrix(value)
        self._dim = value.shape[0]
        self._structure = value
        self._to_rebuild = True

    @property
    def family(self):
        return self._family

    @family.setter
    def family(self, value):
        check_matrix(value)
        check_family(value)
        assert value.shape[0] == self._dim, \
            AttributeError(
                'Family matrix should be of dimension == %d' % (self._dim))
        self._family = value
        self._to_rebuild = True

    @property
    def param1(self):
        return self._param1

    @param1.setter
    def param1(self, value):
        check_matrix(value)
        assert value.shape[0] == self._dim, \
            AttributeError(
                'Family matrix should be of dimension == %d' % (self._dim))
        self._param1 = value
        self._to_rebuild = True

    @property
    def param2(self):
        return self._param2

    @param2.setter
    def param2(self, value):
        if value is None:
            value = np.zeros((self._dim, self._dim))
        check_matrix(value)
        assert value.shape[0] == self._dim, \
            AttributeError(
                'Family matrix should be of dimension == %d' % (self._dim))
        self._param2 = value
        self._to_rebuild = True

    def build_vine(self):
        """
        """
        r_structure = numpy2ri(self.structure)
        r_family = numpy2ri(permute_params(self.family, self.structure))
        r_par = numpy2ri(permute_params(self.param1, self.structure))
        r_par2 = numpy2ri(permute_params(self.param2, self.structure))
        self._rvine = VINECOPULA.RVineMatrix(
            r_structure, r_family, r_par, r_par2)
        self._to_rebuild = False

    def get_sample(self, n_obs):
        """
        """
        assert isinstance(n_obs, int), \
            TypeError("Sample size must be an integer.")
        assert n_obs > 0, \
            ValueError("Sample size must be positive.")
        if self._to_rebuild:
            self.build_vine()
        return np.asarray(VINECOPULA.RVineSim(n_obs, self._rvine))

    def loglikelihood(self, sample):
        """
        """
        data = numpy2ri(sample)
        return np.asarray(VINECOPULA.RVineLogLik(data, self._rvine, separate=True, calculate_V=False)[0])

    def grad_loglikelihood(self, sample):
        """
        """
        data = numpy2ri(sample)
        return np.asarray(VINECOPULA.RVineGrad(data, self._rvine)[0])**2


def permute_params(params, structure):
    """Permute the parameters of the initiat parameter matrix to fit to the R-vine structure.

    Parameters
    ----------
    params : array,
        The matrix of parameters.
    structure : array,
        The R-vine structure array.


    Returns
    -------
    permuted_params : array,
        The permuter matrix of parameters.
    """
    dim = params.shape[0]
    permuted_params = np.zeros(params.shape)
    for i in range(dim):
        for j in range(i+1, dim):
            if structure[i, i] > structure[j, i]:
                coords = structure[i, i]-1, structure[j, i]-1
            else:
                coords = structure[j, i]-1, structure[i, i]-1
            permuted_params[j, i] = params[coords]

    return permuted_params



"""
"""
from .conservative import ConservativeEstimate, ListDependenceResult
from .iterative_vines import iterative_vine_minimize
#except ImportError:
#    print("No module OpenTURNS... Maybe you don't need it!")
from .utils import proba_func, quantile_func
from .vinecopula import VineCopula

__all__ = ["ConservativeEstimate", "proba_func", "quantile_func", 
           "iterative_vine_minimize", "VineCopula", "ListDependenceResult"]



import numpy as np
import random

import openturns as ot
from pyDOE import lhs

from .conversion import Conversion


"""
TODO :
    - Make the classes in Cython for better performances??
    - Change the functions to create a sample for custom number of correlation inputs
    and not conditionally to the parameters dimension.
"""


class SemiDefiniteMatrixError(Exception):
    """
    Specific exception for matrix which are not semi-definite positives
    """
    pass


def create_corr_matrix(rho, dim=None, library="openturns"):
    """
    Create correlation matrices from a list of input parameters
    """
    rho = np.asarray(rho)  # We convert the parameter in numpy array

    # Check if the correlation parameter is correct
    if (rho >= 1.).any() or (rho <= -1.).any():
        raise ValueError("Correlation parameters not in [-1., 1.]")

    # Dimenion of correlation parameter
    corr_dim = len(rho)

    # The dimension problem is computed if not specified
    if not dim:
        root = np.roots([1., -1., -2.*corr_dim])[0]
        dim = int(root)

    # Initialize the matrix
    if library == "openturns":
        corr_matrix = ot.CorrelationMatrix(dim)
    elif library == "numpy":
        corr_matrix = np.identity(dim, dtype=np.float)
    else:
        raise ValueError("Unknow value for library")

    k = 0
    for i in range(dim):
        for j in range(i+1, dim):
            corr_matrix[i, j] = rho[k]
            if library == "numpy":
                corr_matrix[j, i] = rho[k]
            k += 1

    # Check if the matrix is Positive Definite
    error = SemiDefiniteMatrixError("The matrix is not semi-positive definite")
    if library == "openturns":
        if corr_matrix.isPositiveDefinite():
            raise error
    elif library == "numpy":
        if (np.linalg.eigvals(corr_matrix) <= 0.).any():
            raise error

    return corr_matrix


def check_params(rho, dim=None):
    """
    Check is the matrix of a given 
    """
    rho = np.asarray(rho)  # We convert the parameter in numpy array

    # Check if the correlation parameter is correct
    if (rho >= 1.).any() or (rho <= -1.).any():
        raise ValueError("Correlation parameters not in [-1., 1.]")

    # Dimenion of correlation parameter
    corr_dim = len(rho)

    # The dimension problem is computed if not specified
    if not dim:
        root = np.roots([1., -1., -2.*corr_dim])[0]
        dim = int(root)

    # Initialize the matrix
    corr_matrix = ot.CorrelationMatrix(dim)

    k = 0
    for i in range(dim):
        for j in range(i+1, dim):
            corr_matrix[i, j] = rho[k]
            k += 1

    return corr_matrix.isPositiveDefinite()


def create_random_correlation_param_previous(dim, n=1, sampling="monte-carlo"):
    """
    Using acceptation reject...
    """
    # Dimenion of correlation parameter
    corr_dim = dim * (dim - 1) / 2

    # Array of correlation parameters
    list_rho = np.zeros((n, corr_dim), dtype=np.float)

    for i in range(n):  # For each parameter
        condition = True
        # Stop when the matrix is definit semi positive
        while condition:
            if sampling == "monte-carlo":
                rho = np.random.uniform(-1., 1., corr_dim)
            elif sampling == "lhs":
                rho = (lhs(corr_dim, samples=1)*2. - 1.).ravel()
            else:
                raise ValueError("Unknow sampling strategy")
            if check_params(rho, dim):
                condition = False
        list_rho[i, :] = rho

    if n == 1:
        return list_rho.ravel()
    else:
        return list_rho


def create_random_kendall_tau(corr_variables, n=1, sampling="monte-carlo"):
    """
    Using acceptation reject...

    corr_variables: the matrix to defined the variables that are correlated
    """
    dim = corr_variables.shape[0]
    corr_dim = dim * (dim - 1) / 2
    corr_vars = []  # Correlated variables
    k = 0
    for i in range(1, dim):
        for j in range(i):
            # If the variables are correlated,
            # we add the correlation ID in the list
            if corr_variables[i, j]:
                corr_vars.append(k)
            k += 1
    n_corr_vars = len(corr_vars)

    # Array of correlation parameters
    list_tau = np.zeros((n, corr_dim))

    for i in range(n):  # For each parameter
        condition = True
        # Stop when the matrix is definite semi positive
        while condition:
            if sampling == "monte-carlo":
                u = np.random.uniform(-1., 1., n_corr_vars)
            elif sampling == "lhs":
                u = (lhs(n_corr_vars, samples=1)*2. - 1.).ravel()
            else:
                raise ValueError("Unknow sampling strategy")

            u = Conversion.NormalCopula.fromKendallToParam(u)

            if n_corr_vars == corr_dim:
                rho = u
            else:
                rho = np.zeros(corr_dim)
                rho[corr_vars] = u

            if check_params(rho, dim):
                condition = False

        list_tau[i, :] = Conversion.NormalCopula.fromParamToKendall(rho)

    if n == 1:
        return list_tau.ravel()
    else:
        return list_tau


def create_random_correlation_param(corr_variables, n=1, sampling="monte-carlo"):
    """
    Using acceptation reject...

    corr_variables: the matrix to defined the variables that are correlated
    """

    # Dimension problem
    dim = corr_variables.shape[0]
    # Number of correlation parameters
    corr_dim = dim * (dim - 1) / 2
    # Correlated variables
    corr_vars = []
    k = 0
    for i in range(dim):
        for j in range(i+1, dim):
            # If the variables are correlated,
            # we add the correlation ID in the list
            if corr_variables[i, j]:
                corr_vars.append(k)
            k += 1
    n_corr_vars = len(corr_vars)
    # Array of correlation parameters
    list_rho = np.zeros((n, corr_dim), dtype=np.float)

    for i in range(n):  # For each parameter
        condition = True
        # Stop when the matrix is definit semi positive
        while condition:
            if sampling == "monte-carlo":
                u = np.random.uniform(-1., 1., n_corr_vars)
            elif sampling == "lhs":
                u = (lhs(n_corr_vars, samples=1)*2. - 1.).ravel()
            else:
                raise ValueError("Unknow sampling strategy")

            if n_corr_vars == corr_dim:
                rho = u
            else:
                rho = np.zeros(corr_dim)
                rho[corr_vars] = u

            if check_params(rho, dim):
                condition = False
        list_rho[i, :] = rho

    if n == 1:
        return list_rho.ravel()
    else:
        return list_rho


def get_random_rho_3d(size, dim, rho_min=-1., rho_max=1.):
    """
    Works in 1 and 3 dimension
    """
    if dim == 1:
        list_rho = np.asarray(ot.Uniform(rho_min, rho_max).getSample(size))
    else:  # TODO : make it available in d dim
        list_rho = np.zeros((size, dim))
        for i in range(size):
            rho1 = random.uniform(rho_min, rho_max)
            rho2 = random.uniform(rho_min, rho_max)
            l_bound = rho1*rho2 - np.sqrt((1-rho1**2)*(1-rho2**2))
            u_bound = rho1*rho2 + np.sqrt((1-rho1**2)*(1-rho2**2))
            rho3 = random.uniform(l_bound, u_bound)
            list_rho[i, :] = [rho1, rho2, rho3]

    return list_rho


def get_list_rho3(rho1, rho2, n):
    """
    """
    l_bound = rho1*rho2 - np.sqrt((1-rho1**2)*(1-rho2**2))
    u_bound = rho1*rho2 + np.sqrt((1-rho1**2)*(1-rho2**2))
    list_rho3 = np.linspace(l_bound, u_bound, n+1, endpoint=False)[1:]
    return list_rho3


def get_grid_rho(corr_variables, n):
    """
    TODO: think about how to build a sample when not all the variables
    Get the grid of rho parameters
    """
    # Dimension problem
    dim = corr_variables.shape[0]
    # Number of correlation parameters
    corr_dim = dim * (dim - 1) / 2
    # Correlated variables
    corr_vars = []
    k = 0
    for i in range(dim):
        for j in range(i+1, dim):
            # If the variables are correlated,
            # we add the correlation ID in the list
            if corr_variables[i, j]:
                corr_vars.append(k)
            k += 1

    n_corr_vars = len(corr_vars)
    # Array of correlation parameters
    list_rho = np.zeros((n, corr_dim), dtype=np.float)

    if n_corr_vars == 1:  # Easy
        grid = np.linspace(-1., 1., n + 1, endpoint=False)[1:].reshape(n, 1)
    else:  # Not that easy...
        raise NotImplementedError('Not implemented for dim > 1')

    list_rho[:, corr_vars] = grid

    return list_rho


def get_grid_rho_prev(n_sample, dim=3, corr_dim=None, all_sample=True):
    """
    TODO: think about how to build a sample when not all the variables are correlated.
    Get the grid of rho parameters
    """
    if not corr_dim:
        corr_dim = dim * (dim - 1) / 2
    if all_sample:
        n = int(np.floor(n_sample**(1./corr_dim)))
    else:
        n = n_sample

    if corr_dim == 1:  # Easy
        return np.linspace(-1., 1., n + 1, endpoint=False)[1:]
    else:  # Not that easy...
        v_rho = [np.linspace(-1., 1., n+1, endpoint=False)[1:]]*(corr_dim - 1)
        grid_rho = np.meshgrid(*v_rho)
        list_rho = np.vstack(grid_rho).reshape(corr_dim - 1, -1).T

        list_rho = np.zeros((n**corr_dim, corr_dim))
        for i, rho_1 in enumerate(list_rho_1):
            a1 = rho_1[0]
            a2 = rho_1[1]
            list_rho_2 = get_list_rho3(a1, a2, n)
            for j, rho_2 in enumerate(list_rho_2):
                tmp = rho_1.tolist()
                tmp.append(rho_2)
                list_rho[n*i+j, :] = tmp

        return list_rho



from collections import Counter

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy import stats

from .utils import get_grid_sample, to_copula_params

sns.set(style="ticks", color_codes=True)


def set_style_paper():
    # This sets reasonable defaults for font size for
    # a figure that will go in a paper
    sns.set_context("paper")

    # Set the font to be serif, rather than sans
    sns.set(font='serif')

    # Make the background white, and specify the
    # specific font family
    sns.set_style("white", {
        "font.family": "serif",
        "font.serif": ["Times", "Palatino", "serif"]
    })


def get_all_quantity(results, q_func=None):
    """
    """
    quantities = []
    for res_name in results:
        if q_func is not None:
            # We change the quantity function
            results[res_name].q_func = q_func
        min_quantity = results[res_name].min_quantity
        quantities.append(min_quantity)
    return quantities


def get_all_min_result(results, q_func=None):
    """
    """
    min_results = []
    for res_name in results:
        if q_func is not None:
            # We change the quantity function
            results[res_name].q_func = q_func
        min_result = results[res_name].min_result
        min_results.append(min_result)
    return min_results


def get_min_result(all_min_results, q_func=None):
    """
    """
    min_result = None
    min_quantity = np.inf
    for result in all_min_results:
        if q_func is not None:
            # We change the quantity function
            result.q_func = q_func
        if result.min_quantity < min_quantity:
            min_result = result.min_result
            min_quantity = result.min_quantity

    return min_result


def get_n_pairs(all_results):
    """Get the number of pairs in each experiments of an dictionary of iterative results.
    """
    n_pairs = []
    for results in all_results:
        n_pairs_result = Counter()
        for res_name in results:
            n_pair = results[res_name].n_pairs
            n_pairs_result[n_pair] += 1
        assert len(n_pairs_result) == 1, "Not the same number of pairs... Weird"
        n_pairs.append(n_pair)
    return n_pairs


def corrfunc_plot(x, y, **kws):
    """


    Source: https://stackoverflow.com/a/30942817/5224576
    """
    r, _ = stats.pearsonr(x, y)
    k, _ = stats.kendalltau(x, y)
    ax = plt.gca()
    ax.annotate("r = {:.2f}\nk = {:.2f}".format(r, k),
                xy=(.1, .8), xycoords=ax.transAxes,
                weight='heavy', fontsize=14)


def matrix_plot_input(result, kde=False, margins=None):
    """
    """
    input_sample = result.input_sample

    if margins:
        sample = np.zeros(input_sample.shape)
        for i, marginal in enumerate(margins):
            for j, ui in enumerate(input_sample[:, i]):
                sample[j, i] = marginal.computeCDF(ui)
    else:
        sample = input_sample

    data = pd.DataFrame(sample)
    plot = sns.PairGrid(data, palette=["red"])
    if kde:
        plot.map_upper(plt.scatter, s=10)
        plot.map_lower(sns.kdeplot, cmap="Blues_d")
    else:
        plot.map_offdiag(plt.scatter, s=10)

    plot.map_diag(sns.distplot, kde=False)
    plot.map_lower(corrfunc_plot)

    if margins:
        plot.set(xlim=(0, 1), ylim=(0, 1))

    return plot


def matrix_plot_quantities(results, indep_result=None, grid_result=None,
                           q_func=None, figsize=(9, 7), dep_measure='kendalls',
                           quantity_name='Quantity', with_bootstrap=False):
    """
    """
    if isinstance(results, dict):
        input_dim = list(results.values())[0].input_dim
    else:
        input_dim = results.input_dim

    # Figure
    fig, axes = plt.subplots(input_dim, input_dim,
                             figsize=figsize, sharex=True, sharey=True)
    for res in results:
        t = res.split(', ')[-1:-3:-1]
        i = int(t[1][1])
        j = int(t[0][0])
        ax = axes[i, j]
        if dep_measure == 'dependence-param':
            measure = results[res].dep_params
        else:
            measure = results[res].kendalls
        quantities = results[res].quantities
        ax.plot(measure, quantities, '.')

    if dep_measure == 'dependence-param':
        x_label = 'Dependence Parameter'
    else:
        x_label = 'Kendall tau'
    for i in range(input_dim):
        axes[i, 0].set_ylabel(quantity_name)
        axes[-1, i].set_xlabel(x_label)

    fig.tight_layout()


def plot_iterative_results(iter_results, indep_result=None, grid_results=None, q_func=None, figsize=(8, 4),
                           quantity_name='Quantity', with_bootstrap=False, n_boot=200, ax=None):
    """
    """

    # Figure
    if ax is None:
        _, ax = plt.subplots(figsize=figsize)

    # Number of trees
    n_levels = iter_results.iteration+1
    dim = iter_results.dim

    # Colors of the levels and independence
    cmap = plt.get_cmap('jet')

    n_p = 0  # Number of additional plots
    n_p += 1 if indep_result is not None else 0
    n_p += 1 if grid_results is not None else 0
    colors = [cmap(i) for i in np.linspace(0, 1, n_levels+n_p)]

    # Number of pairs at each iteration
    n_pairs = range(1, n_levels+1)

    if indep_result is not None:
        ax.plot([n_pairs[0], n_pairs[-1]], [indep_result.quantity]*2, '-o',
                color=colors[0], label='independence')

        if with_bootstrap:
            indep_result.compute_bootstrap()
            boot = indep_result.bootstrap_sample

            up = np.percentile(boot, 99)
            down = np.percentile(boot, 1)
            ax.plot([n_pairs[0], n_pairs[-1]], [up]*2, '--',
                    color=colors[0], linewidth=0.8)
            ax.plot([n_pairs[0], n_pairs[-1]], [down]*2, '--',
                    color=colors[0], linewidth=0.8)

    if grid_results is not None:
        min_grid_result = grid_results.min_result
        ax.plot([n_pairs[0], n_pairs[-1]], [min_grid_result.quantity]*2, '-o',
                color=colors[1], label='grid-search with $K=%d$' % (grid_results.n_params))
        if with_bootstrap:
            min_grid_result.compute_bootstrap()
            boot = min_grid_result.bootstrap_sample
            up = np.percentile(boot, 95)
            down = np.percentile(boot, 5)
            ax.plot([n_pairs[0], n_pairs[-1]], [up]*2, '--',
                    color=colors[1], linewidth=0.8)
            ax.plot([n_pairs[0], n_pairs[-1]], [down]*2, '--',
                    color=colors[1], linewidth=0.8)

    quantities = []
    min_results_level = []
    for i in range(n_levels):
        values = iter_results.min_quantities(i)[np.tril_indices(dim, -1)]
        values = values[values != 0.].tolist()
        quantities.append(values)
        min_results_level.append(iter_results.min_result(i))

    # Get the minimum of each level
    min_quantities = []
    for quant_lvl in quantities:
        min_quant = min(quant_lvl)
        min_quantities.append(min_quant)

        # Remove the minimum from the list of quantities
        quant_lvl.remove(min_quant)

    for lvl in range(n_levels):
        # The quantities of this level
        quant_lvl = np.asarray(quantities[lvl])
        # The number of results
        n_res = len(quant_lvl)
        ax.plot([n_pairs[lvl]]*n_res, quant_lvl, '.', color=colors[lvl+n_p])

    for lvl in range(n_levels):
        if n_pairs[lvl] == n_pairs[-1]:
            ax.plot(n_pairs[lvl], min_quantities[lvl],
                    'o', color=colors[lvl+n_p])
            if with_bootstrap:
                min_results_level[lvl].compute_bootstrap(n_boot)
                boot = min_results_level[lvl].bootstrap_sample
                up = np.percentile(boot, 95)
                down = np.percentile(boot, 5)
                ax.plot(n_pairs[lvl], up, '.',
                        color=colors[lvl+n_p], linewidth=0.8)
                ax.plot(n_pairs[lvl], down, '.',
                        color=colors[lvl+n_p], linewidth=0.8)
        else:
            ax.plot([n_pairs[lvl], n_pairs[lvl+1]],
                    [min_quantities[lvl]]*2, 'o-', color=colors[lvl+n_p])
            if with_bootstrap:
                min_results_level[lvl].compute_bootstrap(n_boot)
                boot = min_results_level[lvl].bootstrap_sample
                up = np.percentile(boot, 95)
                down = np.percentile(boot, 5)
                ax.plot([n_pairs[lvl], n_pairs[lvl+1]], [up]*2, '--',
                        color=colors[lvl+n_p], linewidth=0.8)
                ax.plot([n_pairs[lvl], n_pairs[lvl+1]], [down]*2, '--',
                        color=colors[lvl+n_p], linewidth=0.8)

    ax.axis('tight')
    ax.set_xlabel('Number of considered pairs')
    ax.set_ylabel(quantity_name)
    ax.set_xticks(n_pairs)
    ax.legend(loc=0)


def compute_influence(obj, K, n, copulas, pair, eps=1.E-4):
    """
    """
    kendalls_fixed = [[]]*2
    bounds = [[eps, 1.-eps]]
    kendalls_fixed[0] = get_grid_sample(bounds, K, 'lhs')
    bounds = [[-1.+eps, -eps]]
    kendalls_fixed[1] = get_grid_sample(bounds, K, 'lhs')

    families = np.zeros((obj.families.shape), dtype=int)
    families[pair[0], pair[1]] = 1
    obj.families = families
    indep_output_sample = obj.independence(n).output_sample
    perfect_output_sample = obj.gridsearch(None, n, 'vertices').output_samples

    output_samples = {}
    for copula in copulas:
        res_out_samples = []
        res_kendalls = []
        for i, num in enumerate(copulas[copula]):
            families[pair[0], pair[1]] = num
            obj.families = families
            converter = [obj._copula_converters[k] for k in obj._pair_ids]
            params = to_copula_params(converter, kendalls_fixed[i])
            output_sample = obj.run_stochastic_models(
                params, n, return_input_sample=False)[0]
            res_out_samples.append(np.asarray(output_sample))

        output_samples[copula] = np.r_[np.concatenate(
            res_out_samples), indep_output_sample.reshape(1, -1), perfect_output_sample].T

    kendalls = np.concatenate(kendalls_fixed).ravel()
    kendalls = np.r_[kendalls, 0.]
    kendalls = np.r_[kendalls, -1., 1.]

    return kendalls, output_samples


def plot_variation(output_samples, kendalls, q_func, plot_area='left', plt_lib='seaborn', figsize=(7, 4), ylabel=None,
                   colors={'Normal': 'b', 'Clayton': 'g', 'Gumbel': 'r', 'Joe': 'm'}, n_boot=5000, ci=99.9):
    """
    """
    set_style_paper()

    if plot_area == 'full':
        taken = np.ones(kendalls.shape, dtype=bool)
    elif plot_area == 'left':
        taken = kendalls <= 0.
    elif plot_area == 'right':
        taken = kendalls >= 0.

    sorting = np.argsort(kendalls[taken])
    fig, ax = plt.subplots(figsize=figsize)

    for copula in output_samples:
        if plt_lib == 'matplotlib':
            quantities = q_func(output_samples[copula].T)
            ax.plot(kendalls[taken][sorting], quantities[taken]
                    [sorting], 'o-', label=copula, markersize=5)
        else:
            sns.tsplot(output_samples[copula][:, taken], time=kendalls[taken],
                       condition=copula, err_style='ci_band', ci=ci, estimator=q_func,
                       n_boot=n_boot, color=colors[copula], ax=ax)

    ax.set_xlabel('Kendall coefficient')
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    else:
        ax.set_ylabel('Output quantity')
    ax.legend(loc=0)
    ax.axis('tight')
    fig.tight_layout()



"""Impact of Dependencies.

The main class inspect the impact of correlation on a quantity
of interest of a model output.

TODO:
    - Make test functions
    - Clean the code
    - Add a load/save to DependenceResult class
    - User np.tril to take the values of the matrices
"""

import json
import operator
import os
import time
import warnings

import h5py
import numpy as np
import openturns as ot
import pandas as pd
from scipy.stats import gaussian_kde, norm
from sklearn.utils import check_random_state

from .conversion import Conversion, get_tau_interval
from .utils import (asymptotic_error_quantile, bootstrap, dict_to_margins,
                    get_grid_sample, get_pair_id, get_pairs_by_levels,
                    get_possible_structures, list_to_matrix,
                    load_dependence_grid, margins_to_dict, matrix_to_list,
                    save_dependence_grid, to_copula_params, to_kendalls)
from .vinecopula import VineCopula, check_matrix

OPERATORS = {">": operator.gt, ">=": operator.ge,
             "<": operator.lt, "<=": operator.le}

GRID_TYPES = ['lhs', 'rand', "fixed", 'vertices']
DEP_MEASURES = ['kendall', 'parameter']


class ConservativeEstimate(object):
    """
    Conservative estimation, toward dependencies, of a quantity of interest at 
    the output of a computational model.

    In a problem with incomplete dependence information, one can try to 
    determine the worst case scenario of dependencies according to a certain 
    risk. The dependence structure is modeled using parametric copulas. For
    multidimensional problems, in addition of Elliptic copulas, one can use
    R-vines to construct multidimensional copulas.

    Parameters
    ----------
    model_func : callable
        The evaluation model :math:`g : \mathbb R^d \rightarrow \mathbb R`
        such as :math:`Y = g(\mathbf X) \in \mathbb R`.
    margins : list of :class:`~openturns.Distribution`
        The :math:`p` marginal distributions.
    families : :class:`~numpy.ndarray`
        The copula family matrix. It describes the family type of each pair
        of variables. See the Vine Copula package for a description of the
        available copulas and their respective indexes.
    vine_structure : :class:`~numpy.ndarray` or None, optional (default=None)
        The Vine copula structure matrix. It describes the construction of the
        vine tree.
        If None, a default matrix is created.
    fixed_params : :class:`~numpy.ndarray`, str or None, optional(default=None)
        The matrix of copula parameters for the fixed copula. Warning: the 
        matrix should contains NaN for all parameters which are not fixed.
        If str, it should be the path to a csv file describing the matrix.
        If None, no parameters are fixed and a default matrix is created.
    bounds_tau : :class:`~numpy.ndarray`, str or None, optional(default=None)
        The matrix of bounds for the exploration of dependencies. The bounds
        have to be on the Kendall's Tau.
        If str, it should be the path to a csv file describing the matrix.
        If None, no bounds are setted and a default matrix is created.
    copula_type : string, optional (default='vine')
        The type of copula. Available types:
        - 'vine': Vine Copula construction
        - 'normal': Multi dimensionnal Gaussian copula.

    Attributes
    ----------

    """

    def __init__(self,
                 model_func,
                 margins,
                 families,
                 fixed_params=None,
                 bounds_tau=None,
                 vine_structure=None,
                 copula_type='vine'):
        self.model_func = model_func
        self.margins = margins
        self.families = families
        self.bounds_tau = bounds_tau
        self.fixed_params = fixed_params
        self.vine_structure = vine_structure
        self.copula_type = copula_type

    def gridsearch(self,
                   n_dep_param,
                   n_input_sample,
                   grid_type='lhs',
                   dep_measure='kendall',
                   lhs_grid_criterion='centermaximin',
                   keep_input_samples=True,
                   load_grid=None,
                   save_grid=None,
                   use_sto_func=False,
                   random_state=None,
                   verbose=False):
        """Grid search over the dependence parameter space.

        Parameters
        ----------
        n_dep_param : int
            The number of dependence parameters in the grid search.
        n_input_sample : int
            The sample size for each dependence parameter.
        grid_type : 'lhs', 'rand' or 'fixed, optional (default='lhs')
            The type of grid :

            - 'lhs' : a Latin Hypercube Sampling (from pyDOE package) with criterion defined by the parameter lhs_grid_criterion,
            - 'rand' : a random sampling,
            - 'fixed' : an uniform grid,
            - 'vertices' : sampling over the vertices of the space.
        dep_measure : 'kendall' or 'parameter', optional (default='kendall')
            The space in which the dependence parameters are created.
        lhs_grid_criterion : string, optional (default = 'centermaximin')
            Configuration of the LHS grid sampling:

            - 'centermaximin' (default),
            - 'center',
            - 'maximin',
            - 'correlation'.

        Returns
        -------
        A list of DependenceResult instances.

        """
        assert isinstance(
            n_input_sample, int), "The sample size should be an integer."
        assert isinstance(grid_type, str), "Grid type should be a string."
        assert n_input_sample > 0, "The sample size should be positive."
        if isinstance(n_dep_param, int):
            assert n_dep_param > 0, "The grid-size should be positive."
        elif n_dep_param is None:
            assert grid_type == 'vertices', "NoneType for n_dep_params only works for vertices type of grids."
        else:
            "The grid-size should be an integer or a NoneType."
        assert grid_type in GRID_TYPES, "Unknow grid type: {}".format(
            grid_type)
        assert isinstance(
            dep_measure, str), "Dependence measure should be a string."
        assert dep_measure in DEP_MEASURES, "Unknow dependence measure: {}".format(
            dep_measure)
        assert isinstance(keep_input_samples,
                          bool), "keep_input_samples should be a bool."

        rng = check_random_state(random_state)

        t_start = time.clock()

        kendalls = None
        grid_filename = None

        if n_dep_param == None and grid_type == 'vertices':
            load_grid = None
            save_grid = None

        # Load a grid
        if load_grid in [None, False]:
            bounds = self._bounds_par_list
            if verbose:
                print('Time taken:', time.clock())
                print('Creating grid')
                t_start = time.time()
            values = get_grid_sample(bounds, n_dep_param, grid_type)
            n_dep_param = len(values)
            if dep_measure == "parameter":
                params = values
            elif dep_measure == "kendall":
                kendalls = values
                if verbose:
                    print('Time taken:', time.clock())
                    print('Converting to kendall parameters')
                    t_start = time.time()
                converter = [self._copula_converters[k]
                             for k in self._pair_ids]
                params = to_copula_params(converter, kendalls)
            elif dep_measure == 'pearson':
                raise NotImplementedError(
                    'Dependence measure not yet implemented.')
            elif dep_measure == 'spearman':
                raise NotImplementedError(
                    'Dependence measure not yet implemented.')
            else:
                raise ValueError("Unknow dependence measure type")
        else:
            # TODO: correct the loading
            # Load the sample from file and get the filename
            kendalls, grid_filename = load_dependence_grid(
                dirname=grid_path,
                n_pairs=self._n_pairs,
                n_params=n_dep_param,
                bounds_tau=self._bounds_tau_list,
                grid_type=grid_type,
                use_grid=use_grid)
            converter = [self._copula_converters[k] for k in self._pair_ids]
            params = to_copula_params(converter, kendalls)

        # TODO: correct the saving
        # The grid is save if it was asked and if it does not already exists
        if save_grid not in [None, False] and load_grid in [None, False]:
            if kendalls is None:
                kendalls = to_kendalls(self._copula_converters, params)
            grid_filename = save_dependence_grid(grid_path, kendalls, self._bounds_tau_list,
                                                 grid_type)

        # Evaluate the sample
        if verbose:
            print('Time taken:', time.clock())
            print('Sample evaluation')

        # Use less memory
        if use_sto_func:
            output_samples = []
            input_samples = None if not keep_input_samples else []
            for i, param in enumerate(params):
                result = self.stochastic_function(param, n_input_sample,
                                                  return_input_sample=keep_input_samples)
                if keep_input_samples:
                    output_sample, input_sample = result
                    input_samples.append(input_sample)
                else:
                    output_sample = result

                output_samples.append(output_sample)

                if verbose:
                    if n_dep_param > 10:
                        if i % int(n_dep_param/10) == 0:
                            print('Time taken:', time.clock())
                            print('Iteration %d' % (i))
                    else:
                        print('Time taken:', time.clock())
                        print('Iteration %d' % (i))
        else:
            if keep_input_samples:
                output_samples, input_samples = self.run_stochastic_models(
                    params, n_input_sample, return_input_samples=keep_input_samples)
            else:
                output_samples = self.run_stochastic_models(
                    params, n_input_sample, return_input_samples=keep_input_samples)
                input_samples = None

        return ListDependenceResult(margins=self.margins,
                                    families=self.families,
                                    vine_structure=self.vine_structure,
                                    bounds_tau=self.bounds_tau,
                                    fixed_params=self.fixed_params,
                                    dep_params=params,
                                    input_samples=input_samples,
                                    output_samples=output_samples,
                                    run_type='grid-search',
                                    grid_type=grid_type,
                                    random_state=rng,
                                    lhs_grid_criterion=lhs_grid_criterion,
                                    grid_filename=grid_filename)

    def run_stochastic_models(self,
                              params,
                              n_input_sample,
                              return_input_samples=True,
                              random_state=None,
                              verbose=False):
        """This function considers the model output as a stochastic function by 
        taking the dependence parameters as inputs.

        Parameters
        ----------
        params : list, or `np.ndarray`
            The list of parameters associated to the predefined copula.
        n_input_sample : int, optional (default=1)
            The number of evaluations for each parameter
        random_state : 
        """
        check_random_state(random_state)

        # Get all the input_sample
        if verbose:
            print('Time taken:', time.clock())
            print('Creating the input samples')

        input_samples = []
        for param in params:
            full_param = np.zeros((self._corr_dim, ))
            full_param[self._pair_ids] = param
            full_param[self._fixed_pairs_ids] = self._fixed_params_list
            intput_sample = self._get_sample(full_param, n_input_sample)
            input_samples.append(intput_sample)

        if verbose:
            print('Time taken:', time.clock())
            print('Evaluate the input samples')

        # Evaluate the through the model
        outputs = self.model_func(np.concatenate(input_samples))
        # List of output sample for each param
        output_samples = np.split(outputs, len(params))

        if verbose:
            print('Time taken:', time.clock())
        if return_input_samples:
            return output_samples, input_samples
        else:
            return output_samples

    def stochastic_function(self,
                            param,
                            n_input_sample=1,
                            return_input_sample=True,
                            random_state=None):
        """This function considers the model output as a stochastic function by 
        taking the dependence parameters as inputs.

        Parameters
        ----------
        param : float, list, or `np.ndarray`
            The parameters associated to the predefined copula.
        n_input_sample : int, optional (default=1)
            The number of evaluations.
        random_state : 
        """
        check_random_state(random_state)

        if isinstance(param, list):
            param = np.asarray(param)
        elif isinstance(param, float):
            param = np.asarray([param])

        assert param.ndim == 1, 'Only one parameter at a time for the moment'

        full_param = np.zeros((self._corr_dim, ))
        full_param[self._pair_ids] = param
        full_param[self._fixed_pairs_ids] = self._fixed_params_list
        input_sample = self._get_sample(full_param, n_input_sample)

        output_sample = self.model_func(input_sample)

        if return_input_sample:
            return output_sample, input_sample
        else:
            return output_sample

    def incomplete(self, n_input_sample, q_func=np.var,
                   keep_input_sample=True, random_state=None):
        """Simulates from the incomplete probability distribution.

        Parameters
        ----------
        n_input_sample : int
            The number of observations in the sampling of :math:`\mathbf X`.

        Returns
        -------
        """
        rng = check_random_state(random_state)
        assert callable(q_func), "Quantity function is not callable"

        param = [0.]*self._n_pairs
        out = self.stochastic_function(param=param,
                                       n_input_sample=n_input_sample,
                                       return_input_sample=keep_input_sample,
                                       random_state=rng)

        if keep_input_sample:
            output_sample, input_sample = out
        else:
            output_sample = out

        return DependenceResult(margins=self._margins,
                                families=self.families,
                                vine_structure=self.vine_structure,
                                fixed_params=self.fixed_params,
                                input_sample=input_sample,
                                output_sample=output_sample,
                                q_func=q_func,
                                random_state=rng)

    def independence(self, n_input_sample, keep_input_sample=True,
                     random_state=None):
        """Generates and evaluates observations at the independence
        configuration.

        Parameters
        ----------
        n_input_sample : int
            The number of observations in the sampling of :math:`\mathbf{X}`.

        Returns
        -------
        """
        rng = check_random_state(random_state)

        assert isinstance(
            n_input_sample, int), "The sample size should be an integer."
        assert n_input_sample > 0, "The sample size should be positive."

        # Creates the sample of input parameters
        tmp = ot.ComposedDistribution(self._margins).getSample(n_input_sample)
        input_sample = np.asarray(tmp)
        output_sample = self.model_func(input_sample)

        if not keep_input_sample:
            input_sample = None

        return DependenceResult(margins=self._margins,
                                input_sample=input_sample,
                                output_sample=output_sample,
                                random_state=rng)

    def _get_sample(self, param, n_sample, param2=None):
        """Creates the observations of the joint input distribution.

        Parameters
        ----------
        param : :class:`~numpy.ndarray`
            A list of :math:`p` copula dependence parameters.
        n_sample : int
            The number of observations.
        param2 : :class:`~numpy.ndarray`, optional (default=None)
            The 2nd copula parameters. For some copula families
            (e.g. Student)
        """
        dim = self._input_dim
        matrix_param = list_to_matrix(param, dim)

        if self._copula_type == 'vine':
            # TODO: One param is used. Do it for two parameters copulas.
            vine_copula = VineCopula(self._vine_structure, self._families,
                                     matrix_param)
            # Sample from the copula
            # The reshape is in case there is only one sample (for RF tests)
            cop_sample = vine_copula.get_sample(
                n_sample).reshape(n_sample, dim)
        elif self._copula_type == 'normal':
            # Create the correlation matrix
            cor_matrix = matrix_param + matrix_param.T + np.identity(dim)
            cop = ot.NormalCopula(ot.CorrelationMatrix(cor_matrix))
            cop_sample = np.asarray(cop.getSample(n_sample), dtype=float)
        else:
            raise AttributeError('Unknown type of copula.')

        # Applied the inverse transformation to get the sample of the joint distribution
        # TODO: this part is pretty much time consuming...
        input_sample = np.zeros((n_sample, dim))
        for i, quantile_func in enumerate(self._margins_quantiles):
            input_sample[:, i] = np.asarray(
                quantile_func(cop_sample[:, i])).ravel()

        return input_sample

    @property
    def model_func(self):
        """The callable model function.
        """
        return self._model_func

    @model_func.setter
    def model_func(self, func):
        """
        """
        assert callable(func), TypeError(
            "The model function must be callable.")
        self._model_func = func

    @property
    def margins(self):
        """The marginal distributions. 

        List of :class:`~openturns.Distribution` objects.
        """
        return self._margins

    @margins.setter
    def margins(self, margins):
        assert isinstance(margins, (list, tuple)), \
            TypeError("It should be a sequence of OT distribution objects.")

        self._margins_quantiles = []
        for marginal in margins:
            assert isinstance(marginal, ot.DistributionImplementation), \
                TypeError("Must be an OpenTURNS distribution objects.")
            self._margins_quantiles.append(marginal.computeQuantile)

        self._margins = margins
        self._input_dim = len(margins)
        self._corr_dim = int(self._input_dim * (self._input_dim - 1) / 2)

        if hasattr(self, '_families'):
            if self._families.shape[0] != self._input_dim:
                print("Don't forget to change the family matrix.")
        if hasattr(self, '_vine_structure'):
            if self._vine_structure.shape[0] != self._input_dim:
                # If it was a custom vine
                if self._custom_vine_structure:
                    print("Don't forget to change the R-vine array")
                else:
                    self.vine_structure = None
        if hasattr(self, '_bounds_tau'):
            if self._bounds_tau.shape[0] != self._input_dim:
                # If the user cares about the bounds
                if self._custom_bounds_tau:
                    print("Don't forget to change the bounds matrix")
                else:
                    self.bounds_tau = None
        if hasattr(self, '_fixed_params'):
            if self._fixed_params.shape[0] != self._input_dim:
                if self._custom_fixed_params:
                    print("Don't forget to change the fixed params matrix")

    @property
    def families(self):
        """The copula families.

        Matrix array of shape (dim, dim).
        """
        return self._families

    @families.setter
    def families(self, families):
        # If load from a file
        if isinstance(families, str):
            # It should be a path to a csv file
            # TODO: replace pandas with numpy
            families = pd.read_csv(families, index_col=0).values
        elif isinstance(families, np.ndarray):
            pass
        else:
            raise TypeError("Not a good type for the familie matrix.")

        families = families.astype(int)
        check_matrix(families)  # Check the matrix

        self._families = families

        # The family list values. Event the independent ones
        self._family_list = matrix_to_list(families, op_char='>=')

        # Dpendent pairs
        _, self._pair_ids, self._pairs = matrix_to_list(families, return_ids=True,
                                                        return_coord=True, op_char='>')

        # Independent pairs
        _, self._indep_pairs_ids, self._indep_pairs = matrix_to_list(
            families, return_ids=True, return_coord=True, op_char='==')

        self._n_pairs = len(self._pair_ids)

        self._copula_converters = [Conversion(
            family) for family in self._family_list]

        if hasattr(self, '_input_dim'):
            if self._families.shape[0] != self._input_dim:
                print("Don't forget to change the margins.")

        if hasattr(self, '_vine_structure'):
            if self._families.shape[0] != self._vine_structure.shape[0]:
                if self._custom_vine_structure:
                    print("Don't forget to change the R-vine array.")

            # It should always be done in case if a pair has been set independent
            if not self._custom_vine_structure:
                self.vine_structure = None

        if hasattr(self, '_fixed_params'):
            if self._families.shape[0] != self._fixed_params.shape[0]:
                if self._custom_fixed_params:
                    print("Don't forget to change the fixed parameters.")
                else:
                    self.fixed_params = None

        if hasattr(self, '_bounds_tau'):
            if self._families.shape[0] != self._fixed_params.shape[0]:
                if self._custom_bounds_tau:
                    print("Don't forget to change the fixed parameters.")
                else:
                    self.bounds_tau = None

            # It should always be done in case if a pair has been set independent
            if not self._custom_bounds_tau:
                self.bounds_tau = None

    @property
    def corr_dim(self):
        """The number of pairs.
        """
        return self._corr_dim

    @property
    def pairs(self):
        """The possibly dependent pairs.
        """
        return self._pairs

    @property
    def pair_ids(self):
        """The possibly dependent pairs.
        """
        return self._pair_ids

    @property
    def n_pairs(self):
        """The number of possibly dependent pairs.
        """
        return self._n_pairs

    @property
    def copula_type(self):
        """The type of copula.

        Can be "vine", or gaussian type.
        """
        return self._copula_type

    @copula_type.setter
    def copula_type(self, value):
        assert isinstance(value, str), \
            TypeError('Type must be a string. Type given:', type(value))

        if value == "normal":
            families = self._families
            # Warn if the user added a wrong type of family
            if (families[families > 0] != 1).any():
                warnings.warn(
                    'Some families were not normal and you want an elliptic copula.')

            # Set all families to gaussian
            families[families > 0] = 1
            self.families = families
        self._copula_type = value

    @property
    def vine_structure(self):
        """The R-vine array.
        """
        return self._vine_structure

    @vine_structure.setter
    def vine_structure(self, structure):
        if structure is None:
            listed_pairs = self._indep_pairs + self._fixed_pairs
            dim = self.input_dim
            # TODO : this should be corrected...
            if len(listed_pairs) > 0:
                # if False:
                pairs_iter_id = [get_pair_id(
                    dim, pair, with_plus=False) for pair in listed_pairs]
                pairs_by_levels = get_pairs_by_levels(dim, pairs_iter_id)
                structure = get_possible_structures(dim, pairs_by_levels)[1]
            else:
                structure = np.zeros((dim, dim), dtype=int)
                for i in range(dim):
                    structure[i, 0:i+1] = i + 1
            self._custom_vine_structure = False
        else:
            check_matrix(structure)
            self._custom_vine_structure = True
        self._vine_structure = structure
        self._vine_structure_list = structure[np.tril_indices_from(structure)]

    @property
    def input_dim(self):
        """The input dimension.
        """
        return self._input_dim

    @property
    def bounds_tau(self):
        """The matrix bound for the kendall's tau.
        """
        return self._bounds_tau

    @property
    def bounds_par(self):
        """The matrix bound for the dependence parameters.
        """
        return self._bounds_par

    @bounds_tau.setter
    def bounds_tau(self, value):
        """Set the upper bound of the Kendall Tau parameter space.

        Parameters
        ----------
        value : :class:`~numpy.ndarray`, str or None
            Matrix of bounds.
        """

        dim = self._input_dim
        self._custom_bounds_tau = True
        # If no bounds given, we take the min and max, depending on the copula family
        if value is None:
            bounds_tau = np.zeros((dim, dim))
            for i, j in self._pairs:
                bounds_tau[i, j], bounds_tau[j, i] = get_tau_interval(
                    self._families[i, j])
            self._custom_bounds_tau = False
        elif isinstance(value, str):
            # It should be a path to a csv file
            bounds_tau = pd.read_csv(value, index_col=0).values
        else:
            bounds_tau = value

        bounds_par = np.zeros(bounds_tau.shape)
        bounds_tau_list = []
        bounds_par_list = []
        for k, p in enumerate(self._pair_ids):
            i, j = self._pairs[k]
            tau_min, tau_max = get_tau_interval(self._family_list[p])

            if not np.isnan(bounds_tau[i, j]):
                tau_min = max(bounds_tau[i, j], tau_min)

            if not np.isnan(bounds_tau[j, i]):
                tau_max = min(bounds_tau[j, i], tau_max)

            bounds_tau_list.append([tau_min, tau_max])
            # Conversion to copula parameters
            param_min = self._copula_converters[p].to_copula_parameter(
                tau_min, 'kendall-tau')
            param_max = self._copula_converters[p].to_copula_parameter(
                tau_max, 'kendall-tau')

            bounds_par[i, j] = tau_min
            bounds_par[j, i] = tau_max
            bounds_par_list.append([param_min, param_max])

        check_matrix(bounds_tau)
        check_matrix(bounds_par)
        self._bounds_tau = bounds_tau
        self._bounds_par = bounds_par
        self._bounds_tau_list = bounds_tau_list
        self._bounds_par_list = bounds_par_list

    @property
    def fixed_params(self):
        """The pairs that are fixed to a given dependence parameter value.
        """
        return self._fixed_params

    @fixed_params.setter
    def fixed_params(self, value):
        # TODO: if it is changed multiple times, it keeps deleting pairs...
        self._custom_fixed_params = True
        if value is None:
            # There is no fixed pairs
            matrix = np.zeros((self._input_dim, self._input_dim), dtype=float)
            matrix[:] = np.nan
            self._custom_fixed_params = False
        elif isinstance(value, str):
            # It should be a path to a csv file
            matrix = pd.read_csv(value, index_col=0).values
        else:
            matrix = value

        # The matrix should be checked
        check_matrix(matrix)

        # The lists only contains the fixed pairs informations
        self._fixed_pairs = []
        self._fixed_pairs_ids = []
        self._fixed_params = matrix
        self._fixed_params_list = []
        k = 0
        # TODO: do it like for the families property
        for i in range(1, self._input_dim):
            for j in range(i):
                if self._families[i, j] > 0:
                    if matrix[i, j] == 0.:
                        warnings.warn(
                            'The parameter of the pair %d-%d is set to 0. Check if this is correct.' % (i, j))
                    if not np.isnan(matrix[i, j]):
                        # The pair is fixed we add it in the list
                        self._fixed_pairs.append([i, j])
                        self._fixed_pairs_ids.append(k)
                        self._fixed_params_list.append(matrix[i, j])
                        # And we remove it from the list of dependent pairs
                        if k in self._pair_ids:
                            self._pair_ids.remove(k)
                            self._pairs.remove([i, j])
                            self._n_pairs -= 1
                        self._bounds_tau[i, j] = np.nan
                        self._bounds_tau[j, i] = np.nan
                k += 1

        if hasattr(self, '_vine_structure'):
            # It should always be done in case if a pair has been set independent
            if not self._custom_vine_structure:
                self.vine_structure = None

        if hasattr(self, '_bounds_tau'):
            # It should always be done in case if a pair has been set independent
            if not self._custom_bounds_tau:
                self.bounds_tau = None
            else:
                if self.bounds_tau.shape[0] != self.input_dim:
                    print("Dont't foget to update the bounds matrix")
                else:
                    bounds_tau = self.bounds_tau
                    # We delete the bounds for the fixed pairs
                    for fixed_pair in self._fixed_pairs:
                        bounds_tau[fixed_pair[0], fixed_pair[1]] = np.nan
                        bounds_tau[fixed_pair[1], fixed_pair[0]] = np.nan
                    self.bounds_tau = bounds_tau


class ListDependenceResult(list):
    """The result from the Conservative Estimation.

    The results gather in the list must have the same configurations: the same
    copula families, vine structure, grid.

    Parameters
    ----------
    margins : list of OpenTURNS distributions
        The OT distributions.
    families : array
        The matrix array of the families.
    vine_structure : array
        The matrix array of the R-vine. If None, it is considered as Gaussian.
    bounds_tau : array,
        The matrix array of the bounds for the dependence parameters.
    dep_param : array
        The dependence parameters.
    input_sample : array
        The input sample.
    output_sample : array
        The output sample.
    q_func : callable or None
        The output quantity of intereset function.
    run_type : str
        The type of estimation: independence, grid-search, iterative, ...
    grid_type : str
        The type of grid use if it was a grid search.
    random_state : int, RandomState or None,
        The random state of the computation.

    """

    def __init__(self,
                 margins=None,
                 families=None,
                 vine_structure=None,
                 bounds_tau=None,
                 fixed_params=None,
                 dep_params=None,
                 input_samples=None,
                 output_samples=None,
                 q_func=None,
                 run_type=None,
                 n_evals=None,
                 grid_type=None,
                 random_state=None,
                 **kwargs):

        self.margins = margins
        self.families = families
        self.vine_structure = vine_structure
        self.bounds_tau = bounds_tau
        self.fixed_params = fixed_params
        self._q_func = q_func
        self.run_type = run_type
        self.grid_type = grid_type
        self.input_dim = len(margins)
        self.corr_dim = int(self.input_dim * (self.input_dim - 1) / 2)

        self.grid_filename = None
        if "grid_filename" in kwargs:
            self.grid_filename = kwargs["grid_filename"]

        self.lhs_grid_criterion = None
        if "lhs_grid_criterion" in kwargs:
            self.lhs_grid_criterion = kwargs["lhs_grid_criterion"]

        self.output_id = 0
        if "output_id" in kwargs:
            self.output_id = kwargs["output_id"]

        if run_type in ['grid-search', 'iterative']:
            assert output_samples is not None, \
                "Add some output sample if you're adding dependence parameters"

            for k, dep_param in enumerate(dep_params):
                input_sample = None if input_samples is None else input_samples[k]
                output_sample = output_samples[k]

                result = DependenceResult(margins=margins,
                                          families=families,
                                          vine_structure=vine_structure,
                                          fixed_params=fixed_params,
                                          dep_param=dep_param,
                                          input_sample=input_sample,
                                          output_sample=output_sample,
                                          q_func=q_func,
                                          random_state=random_state,
                                          output_id=self.output_id)
                self.append(result)

            if output_sample.shape[0] == output_sample.size:
                self.output_dim = 1
            else:
                self.output_dim = output_sample.shape[1]

        elif run_type == 'independence':
            # There is data and we suppose it's at independence or a fixed params
            result = DependenceResult(margins=margins,
                                      families=families,
                                      vine_structure=vine_structure,
                                      fixed_params=fixed_params,
                                      dep_param=0,
                                      input_sample=input_samples,
                                      output_sample=output_samples[0],
                                      q_func=q_func,
                                      random_state=random_state,
                                      output_id=self.output_id)
            self.families = 0
            self.vine_structure = 0
            self.bounds_tau = 0
            self.fixed_params = 0
            self.grid_type = 0
            self.append(result)
            self.output_dim = result.output_dim

        elif run_type == 'incomplete':
            # There is data and we suppose it's at independence or a fixed params
            result = DependenceResult(margins=margins,
                                      families=families,
                                      vine_structure=vine_structure,
                                      fixed_params=fixed_params,
                                      dep_param=0,
                                      input_sample=input_samples,
                                      output_sample=output_samples[0],
                                      q_func=q_func,
                                      random_state=random_state,
                                      output_id=self.output_id)

            self.grid_type = 0
            self.append(result)
            self.output_dim = result.output_dim

        self.rng = check_random_state(random_state)
        self._bootstrap_samples = None

    def __add__(self, results):
        """
        """
        if self.n_params > 0:
            # Assert the results are the same categories
            np.testing.assert_equal(
                self.margins, results.margins, err_msg="Same margins")
            np.testing.assert_array_equal(
                self.families, results.families, err_msg="Different copula families")
            np.testing.assert_array_equal(
                self.vine_structure, results.vine_structure, err_msg="Different copula structures")
            np.testing.assert_array_equal(
                self.bounds_tau, results.bounds_tau, err_msg="Different bounds on Tau")
            np.testing.assert_array_equal(
                self.fixed_params, results.fixed_params, err_msg="Different fixed params")
            np.testing.assert_allclose(
                self.dep_params, results.dep_params, err_msg="Different dependence parameters")
            assert self.run_type == results.run_type, "Different run type"
            assert self.grid_type == results.grid_type, "Different grid type"
            assert self.grid_filename == results.grid_filename, "Different grid type"
            assert self.lhs_grid_criterion == results.lhs_grid_criterion, "Different grid type"

            input_samples = []
            output_samples = []
            for res1, res2 in zip(self, results):
                if res1.input_sample is not None:
                    input_samples.append(
                        np.r_[res1.input_sample, res2.input_sample])
                output_samples.append(
                    np.r_[res1.output_sample, res2.output_sample])

            if len(input_samples) == 0:
                input_samples = None
            new_results = ListDependenceResult(
                margins=self.margins,
                families=self.families,
                vine_structure=self.vine_structure,
                bounds_tau=self.bounds_tau,
                fixed_params=self.fixed_params,
                dep_params=self.dep_params,
                input_samples=input_samples,
                output_samples=output_samples,
                grid_type=self.grid_type,
                q_func=self.q_func,
                run_type=self.run_type,
                grid_filename=self.grid_filename,
                lhs_grid_criterion=self.lhs_grid_criterion,
                output_id=self.output_id)
        return new_results

    def extend(self, value):
        super(ListDependenceResult, self).extend(value)
        self.families = value.families

    @property
    def output_id(self):
        """Id of the output.
        """
        return self._output_id

    @output_id.setter
    def output_id(self, output_id):
        for result in self:
            result.output_id = output_id
        self._output_id = output_id

    @property
    def q_func(self):
        """The quantity function
        """
        return self._q_func

    @q_func.setter
    def q_func(self, q_func):
        assert callable(q_func), "Function must be callable"
        if self.n_params == 0:
            print("There is no data...")
        else:
            for result in self:
                result.q_func = q_func
        self._q_func = q_func

    @property
    def pairs(self):
        """The dependent pairs of the problem.
        """
        if self.families is None:
            print('Family matrix was not defined')
        else:
            return matrix_to_list(self.families)[1]

    @property
    def dep_params(self):
        """The dependence parameters.
        """
        if self.n_params == 0:
            print("There is no data...")
        else:
            return np.asarray([result.dep_param for result in self])

    @property
    def kendalls(self):
        """The Kendall's tau dependence measure.
        """
        if self.n_params == 0:
            print("There is no data...")
        else:
            return np.asarray([result.kendall_tau for result in self])

    @property
    def n_pairs(self):
        """The number of dependente pairs.
        """
        if self.n_params == 0:
            return 0
        else:
            return (self.families > 0).sum()

    @property
    def output_samples(self):
        if self.n_params == 0:
            print("There is no data...")
        else:
            return [result.output_sample for result in self]

    @property
    def input_samples(self):
        if self.n_params == 0:
            print("There is no data...")
        else:
            return [result.input_sample for result in self]

    @property
    def n_input_sample(self):
        """The sample size for each dependence parameter.
        """
        # TODO: maybe not all the samples have the same number of observations...
        if self.n_params == 0:
            return 0
        else:
            return self[0].n_sample

    @property
    def n_evals(self):
        """The total number of observations.
        """
        return self.n_params*self.n_input_sample

    @property
    def n_params(self):
        """The number of dependence parameters.
        """
        return len(self)

    @property
    def quantities(self):
        """The quantity values of each parameters.
        """
        if self.n_params == 0:
            print("There is no data...")
        else:
            return np.asarray([result.quantity for result in self])

    @property
    def min_result(self):
        """The dependence parameter that minimizes the output quantity.
        """
        if self.n_params == 0:
            print("There is no data...")
        else:
            return self[self.quantities.argmin()]

    @property
    def min_quantity(self):
        """The minimum quantity from all the dependence parameters.
        """
        if self.n_params == 0:
            print("There is no data...")
        else:
            return self.quantities.min()

    @property
    def full_dep_params(self):
        """The dependence parameters with the columns from the fixed parameters.
        """
        if self.n_params == 0:
            print("There is no data...")
        else:
            return np.asarray([result.full_dep_params for result in self])

    @property
    def bootstrap_samples(self):
        """The computed bootstrap sample of all the dependence parameters.
        """
        sample = [result._bootstrap_sample for result in self]
        if not any((boot is None for boot in sample)):
            return np.asarray(sample)
        else:
            raise AttributeError('The boostrap must be computed first')

    def compute_bootstraps(self, n_bootstrap=1000, inplace=True):
        """Compute bootstrap of the quantity for each element of the list
        """
        if self.n_params == 0:
            print("There is no data...")
        else:
            for result in self:
                result.compute_bootstrap(n_bootstrap)

            if not inplace:
                return self.bootstrap_samples

    def to_hdf(self, path_or_buf, input_names=[], output_names=[], verbose=False, with_input_sample=True):
        """Write the contained data to an HDF5 file using HDFStore.

        Parameters
        ----------
        path_or_buf : the path (string) or HDFStore object
            The path of the file or an hdf instance.
        input_names : list of strings, optional
            The name of the inputs variables.
        output_names : list of strings, optional
            The name of the outputs.
        """
        filename, extension = os.path.splitext(path_or_buf)
        dirname = os.path.dirname(path_or_buf)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        assert extension in ['.hdf', '.hdf5'], "File extension should be hdf"

        # List of input variable names
        if input_names:
            assert len(input_names) == self.input_dim, \
                AttributeError("Dimension problem for input_names")
        else:
            for i in range(self.input_dim):
                input_names.append("x_%d" % (i + 1))

        # List of output variable names
        if output_names:
            assert len(output_names) == self.output_dim, \
                AttributeError("Dimension problem for output_names")
        else:
            for i in range(self.output_dim):
                output_names.append("y_%d" % (i + 1))

        margin_dict = margins_to_dict(self.margins)

        filename_exists = True
        k = 0
        while filename_exists:
            # If the file has the same run configuration
            try:
                with h5py.File(path_or_buf, 'a') as hdf_store:
                    # If the file already exists and already has data
                    if hdf_store.attrs.keys():
                        # Check the attributes of the file, if it already exists
                        np.testing.assert_allclose(
                            hdf_store['dependence_params'].value, self.dep_params, err_msg="Different dependence parameters")
                        assert hdf_store.attrs['Input Dimension'] == self.input_dim, "Different input dimension"
                        assert hdf_store.attrs['Output Dimension'] == self.output_dim, "Different output dimension"
                        assert hdf_store.attrs['Run Type'] == self.run_type, "Different run type"
                        np.testing.assert_array_equal(
                            hdf_store.attrs['Copula Families'], self.families, err_msg="Different copula families")
                        if 'Fixed Parameters' in hdf_store.attrs.keys():
                            np.testing.assert_array_equal(
                                hdf_store.attrs['Fixed Parameters'], self.fixed_params, err_msg="Different fixed copulas")
                        elif self._fixed_pairs:
                            # Save only if there is no fixed params
                            raise ValueError(
                                'It should not have constraints to be in the same output file.')
                        if 'Bounds Tau' in hdf_store.attrs.keys():
                            np.testing.assert_array_equal(
                                hdf_store.attrs['Bounds Tau'], self.bounds_tau, err_msg="Different bounds on Tau")
                        elif self._fixed_pairs:
                            raise ValueError(
                                'It should not have constraints to be in the same output file.')
                        np.testing.assert_array_equal(
                            hdf_store.attrs['Copula Structure'], self.vine_structure, err_msg="Different vine structures")
                        np.testing.assert_array_equal(
                            hdf_store.attrs['Input Names'], input_names, err_msg="Different Input Names")
                        np.testing.assert_array_equal(
                            hdf_store.attrs['Output Names'], output_names, err_msg="Different output Names")

                        loaded_margin_dict = json.loads(
                            hdf_store.attrs['Margins'])
                        assert all(loaded_margin_dict[str(
                            k)] == margin_dict[k] for k in margin_dict), "Not the same dictionary"

                        if self.run_type == 'grid-search':
                            assert hdf_store.attrs['Grid Type'] == self.grid_type, "Different grid type"
                    else:
                        # We save the attributes in the empty new file
                        hdf_store.create_dataset(
                            'dependence_params', data=self.dep_params)
                        # Margins
                        hdf_store.attrs['Margins'] = json.dumps(
                            margin_dict, ensure_ascii=False)
                        hdf_store.attrs['Copula Families'] = self.families
                        hdf_store.attrs['Copula Structure'] = self.vine_structure
                        hdf_store.attrs['Bounds Tau'] = self.bounds_tau
                        hdf_store.attrs['Grid Size'] = self.n_params
                        hdf_store.attrs['Input Dimension'] = self.input_dim
                        hdf_store.attrs['Output Dimension'] = self.output_dim
                        hdf_store.attrs['Fixed Parameters'] = self.fixed_params
                        hdf_store.attrs['Run Type'] = self.run_type
                        hdf_store.attrs['Input Names'] = input_names
                        hdf_store.attrs['Output Names'] = output_names

                        if self.run_type == 'grid-search':
                            hdf_store.attrs['Grid Type'] = self.grid_type
                            if self.grid_filename is not None:
                                hdf_store.attrs['Grid Filename'] = os.path.basename(
                                    self.grid_filename)
                            if self.grid_type == 'lhs':
                                hdf_store.attrs['LHS Criterion'] = self.lhs_grid_criterion

                    # Check the number of experiments
                    grp_number = 0
                    list_groups = hdf_store.keys()
                    list_groups.remove('dependence_params')
                    list_groups = [int(g) for g in list_groups]
                    list_groups.sort()

                    # If there is already groups in the file
                    if list_groups:
                        grp_number = list_groups[-1] + 1

                    grp = hdf_store.create_group(str(grp_number))
                    for i in range(self.n_params):
                        grp_i = grp.create_group(str(i))
                        grp_i.attrs['n'] = self[i].n_sample
                        grp_i.create_dataset(
                            'output_sample', data=self[i].output_sample)
                        if with_input_sample:
                            grp_i.create_dataset(
                                'input_sample', data=self[i].input_sample)
                    filename_exists = False
            except AssertionError as msg:
                print('File %s has different configurations' % (path_or_buf))
                if verbose:
                    print(str(msg))
                path_or_buf = '%s_num_%d%s' % (filename, k, extension)
                k += 1

        if verbose:
            print('Data saved in %s' % (path_or_buf))

    @classmethod
    def from_hdf(cls, filepath_or_buffer, id_of_experiment='all', output_id=0,
                 with_input_sample=True, q_func=np.var):
        """Loads result from HDF5 file.

        This class method creates an instance of :class:`~ConservativeEstimate` 
        by loading a HDF5 with the saved result of a previous run.

        Parameters
        ----------
        filepath_or_buffer : str
            The path of the file to hdf5 file read.
        id_of_experiment : str or int, optional (default='all')
            The experiments to load. The hdf5 file can gather multiple 
            experiments with the same metadatas. The user can chose to load all
            or one experiments.
        output_id : int, optional (default=0)
            The index of the output if the function output is multidimensional.
        with_input_sample : bool, optional (default=True)
            If False the samples of input observations are not loaded. Input 
            observations are not necessary to compute output quantity of 
            interests.

        Returns
        -------
        obj : :class:`~ImpactOfDependence`
            The Impact Of Dependence instance with the loaded informations.
        """
        # Load of the hdf5 file
        with h5py.File(filepath_or_buffer, 'r') as hdf_store:
            # The file may contain multiple experiments. The user can load one
            # or multiple experiments if they have similiar configurations.
            if id_of_experiment == 'all':
                # All groups of experiments are loaded and concatenated
                list_index = hdf_store.keys()
                list_index.remove('dependence_params')
            else:
                # Only the specified experiment is loaded
                assert isinstance(id_of_experiment, int), 'It should be an int'
                list_index = [str(id_of_experiment)]

            params = hdf_store['dependence_params'].value
            run_type = hdf_store.attrs['Run Type']
            families = hdf_store.attrs['Copula Families']
            vine_structure = hdf_store.attrs['Copula Structure']
            #copula_type = hdf_store.attrs['Copula Type']
            input_dim = hdf_store.attrs['Input Dimension']
            input_names = hdf_store.attrs['Input Names']

            # Many previous experiments did not have this attribute.
            # The checking is temporary and should be deleted in future
            # versions.
            fixed_params = None
            if 'Fixed Parameters' in hdf_store.attrs.keys():
                fixed_params = hdf_store.attrs['Fixed Parameters']
            bounds_tau = None
            if 'Bounds Tau' in hdf_store.attrs.keys():
                bounds_tau = hdf_store.attrs['Bounds Tau']

            margins = dict_to_margins(json.loads(hdf_store.attrs['Margins']))

            grid_type = None
            grid_filename = None
            lhs_grid_criterion = None
            if run_type == 'grid-search':
                grid_type = hdf_store.attrs['Grid Type']
                if 'Grid Filename' in hdf_store.attrs.keys():
                    grid_filename = hdf_store.attrs['Grid Filename']
                if grid_type == 'lhs':
                    lhs_grid_criterion = hdf_store.attrs['LHS Criterion']

            output_names = hdf_store.attrs['Output Names']

            # For each experiment
            for j_exp, index in enumerate(list_index):
                grp = hdf_store[index]  # Group of the experiment

                input_samples = None
                if with_input_sample:
                    input_samples = []

                output_samples = []
                n_samples = []
                elements = [int(i) for i in grp.keys()]
                for k in sorted(elements):
                    res = grp[str(k)]
                    if with_input_sample:
                        data_in = res['input_sample'].value
                    data_out = res['output_sample'].value

                    if with_input_sample:
                        input_samples.append(data_in)
                    output_samples.append(data_out)
                    n_samples.append(res.attrs['n'])

                result = cls(margins=margins,
                             families=families,
                             vine_structure=vine_structure,
                             bounds_tau=bounds_tau,
                             fixed_params=fixed_params,
                             dep_params=params,
                             input_samples=input_samples,
                             output_samples=output_samples,
                             grid_type=grid_type,
                             q_func=q_func,
                             run_type=run_type,
                             grid_filename=grid_filename,
                             lhs_grid_criterion=lhs_grid_criterion,
                             output_id=output_id)

                if j_exp == 0:
                    results = result
                else:
                    results = results + result

        return results


class DependenceResult(object):
    """Result from conservative estimate.

    Parameters
    ----------
    margins : list
        The OT distributions.
    families : array
        The matrix array of the families.
    vine_structure : array
        The matrix array of the R-vine. If None, it is considered as Gaussian.
    dep_param : array
        The dependence parameters.
    input_sample : array
        The input sample.
    output_sample : array
        The output sample.
    q_func : callable or None
        The output quantity of intereset function.
    random_state : int, RandomState or None,
        The random state of the computation.
    """

    def __init__(self,
                 margins=None,
                 families=None,
                 vine_structure=None,
                 fixed_params=None,
                 dep_param=None,
                 input_sample=None,
                 output_sample=None,
                 q_func=None,
                 output_id=0,
                 random_state=None):

        self.margins = margins
        self.families = families
        self.vine_structure = vine_structure
        self.fixed_params = fixed_params
        self.dep_param = dep_param
        self.input_sample = input_sample
        self.output_sample = output_sample
        self.q_func = q_func
        self.rng = check_random_state(random_state)
        self.output_id = output_id

        self.n_sample = output_sample.shape[0]
        self.input_dim = len(margins)
        if output_sample.shape[0] == output_sample.size:
            self.output_dim = 1
        else:
            self.output_dim = output_sample.shape[1]
        self.corr_dim = int(self.input_dim * (self.input_dim - 1) / 2)
        self._bootstrap_sample = None
        self._n_bootstrap_sample = None
        self._gaussian_kde = None

    def compute_bootstrap(self, n_bootstrap=1000, inplace=True):
        """Bootstrap of the output quantity of interest.

        Parameters
        ----------
        n_bootstrap : int, optional
            The number of bootstrap samples.
        inplace : bool, optional
           If true, the bootstrap sample is returned

        Returns
        -------
            The bootstrap sample if inplace is true.
        """
        self._bootstrap_sample = bootstrap(
            self.output_sample_id, n_bootstrap, self.q_func)
        self._n_bootstrap_sample = self._bootstrap_sample.shape[0]
        if (self._bootstrap_sample is None) or (self._n_bootstrap_sample != n_bootstrap):
            self.compute_bootstrap(n_bootstrap)

        self._std = self._bootstrap_sample.std()
        self._mean = self._bootstrap_sample.mean()
        self._cov = abs(self._std/self._mean)

        if not inplace:
            return self._bootstrap_sample

    def compute_quantity_bootstrap_ci(self, alphas=[0.025, 0.975], n_bootstrap=1000):
        """Boostrap confidence interval.
        """
        if (self._bootstrap_sample is None) or (self._n_bootstrap_sample != n_bootstrap):
            self.compute_bootstrap(n_bootstrap)

        return np.percentile(self._bootstrap_sample, [a*100. for a in alphas]).tolist()

    def compute_quantity_asymptotic_ci(self, quantity_name, quantity_param, ci=0.95):
        """Asymptotic confidence interval.
        """
        quantity = self.quantity

        if quantity_name == 'quantile':
            density = self.kde_estimate(self.quantity)[0]
            error = asymptotic_error_quantile(
                self.n_sample, density, quantity_param)
        elif quantity_name == 'probability':
            error = asymptotic_error_quantile(self.n_sample, quantity)
        else:
            raise 'Unknow quantity_name: {0}'.format(quantity_name)
        gaussian_quantile = norm.ppf(1. - (1. - ci)/2.)
        deviation = gaussian_quantile*error
        return [quantity - deviation, quantity + deviation]

    @property
    def boot_cov(self):
        """Coefficient of variation.
        """
        if self._cov is None:
            print('Create a bootstrap sample first.')
        return self._cov

    @property
    def boot_mean(self):
        """Mean of the quantity.
        """
        if self._mean is None:
            print('Create a bootstrap sample first.')
        return self._mean

    @property
    def boot_var(self):
        """Standard deviation of the quantity
        """
        if self._std is None:
            print('Create a bootstrap sample first.')
        return self._std

    @property
    def kde_estimate(self):
        """
        """
        if self._gaussian_kde is not None:
            return self._gaussian_kde
        else:
            self._gaussian_kde = gaussian_kde(self.output_sample_id)
            return self._gaussian_kde

    @property
    def bootstrap_sample(self):
        """The computed bootstrap sample.
        """
        if self._bootstrap_sample is not None:
            return self._bootstrap_sample
        else:
            raise AttributeError('The boostrap must be computed first')

    @property
    def quantity(self):
        """The computed output quantity.
        """
        # TODO: don't compute it everytime...
        quantity = self.q_func(self.output_sample_id, axis=0)
        return quantity.item() if quantity.size == 1 else quantity

    @property
    def output_sample_id(self):
        """
        """
        if self.output_dim == 1:
            return self.output_sample
        else:
            return self.output_sample[:, self.output_id]

    @property
    def full_dep_params(self):
        """The matrix of parameters for all the pairs.
        """
        full_params = np.zeros((self.corr_dim, ))
        pair_ids = matrix_to_list(self.families, return_ids=True)[1]
        full_params[pair_ids] = self.dep_param
        if self.fixed_params is not None:
            fixed_params, fixed_pairs = matrix_to_list(
                self.fixed_params, return_ids=True)
            full_params[fixed_pairs] = fixed_params
        return full_params

    @property
    def kendall_tau(self):
        """The Kendall's tau of the dependence parameters.
        """
        kendalls = []
        for family, id_param in zip(*matrix_to_list(self.families, return_ids=True)):
            kendall = Conversion(family).to_kendall(
                self.full_dep_params[id_param])
            if kendall.size == 1:
                kendall = kendall.item()
            kendalls.append(kendall)
        return kendalls

    @property
    def full_kendall_tau(self):
        """The Kendall's tau of the dependence parameters.
        """
        kendalls = []
        for family, id_param in zip(*matrix_to_list(self.families, return_ids=True, op_char='>=')):
            kendall = Conversion(family).to_kendall(
                self.full_dep_params[id_param])
            if kendall.size == 1:
                kendall = kendall.item()
            kendalls.append(kendall)
        return kendalls



#!/usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='dep-impact',
    version='0.1',
    description='A Python package on the influence of dependencies in a structural reliability problem.',
    url='https://github.com/NazBen/dep-impact',
    author='Nazih Benoumechiara',
    author_email='nazih.benoumechiara@gmail.com',
    license='MIT',
    keywords='copula reliability regular-vines',
    packages=['depimpact'],
    install_requires=required
)
